import numpy as np
import matplotlib.pyplot as plt
import copy
from fconcrete.helpers import getAxis

class LongSteelBar():
    def __init__(self, long_begin, long_end, quantity, quantity_accumulated, diameter, area, area_accumulated, fyd, interspace, length, cost):
        self.long_begin = long_begin
        self.long_end = long_end
        self.quantity = quantity
        self.diameter = diameter
        self.interspace = interspace
        self.quantity_accumulated = quantity_accumulated
        self.area_accumulated = area_accumulated
        self.area = area
        self.fyd = fyd
        self.length = length
        self.cost = cost
    
    @staticmethod
    def getSteelArea(section, material, steel, momentum):
        b = section.width()
        positive_steel_height, negative_steel_height = section.positive_steel_height, section.negative_steel_height
        if momentum==0: return 0
        d = positive_steel_height if momentum>0 else negative_steel_height
        #print(momentum, positive_steel_height, negative_steel_height, d)
        kc = b*d**2/momentum
        if (kc<1.5 and kc>-1.5): raise Exception('Momentum too high to section')
        fyd = steel.fyd
        fcd = material.fcd
        beta_x = (1-(1-1.6/(0.68*fcd*kc))**(0.5))/0.8
        tension_steel = np.where((beta_x <= 0.28) or (3.5*(beta_x**(-1)-1)*20>fyd), fyd, 3.5*(beta_x**(-1)-1)*20)+0
        ks = (tension_steel*(1-0.4*beta_x))**(-1)
        As = ks*momentum/d
        return As
    
    @staticmethod
    def getMinimumAndMaximumSteelArea(area, fck):
        """
            Giving the fck in kN/cmˆ2, returns the minimum and maximum area.
        """
        fck_array = 20 + 5*np.arange(15)
        fck_array = fck_array/10
        p_min_array = np.array([ 0.115, 0.115, 0.115, 0.164, 0.179, 0.194, 0.208, 0.211, 0.219, 0.226, 0.233, 0.239, 0.245, 0.251, 0.256 ])
        p_min = np.interp(fck, fck_array, p_min_array)
        A_min = area*p_min/100
        A_max = 0.008*area
        return A_min, A_max
    
        
    def plot(self,prop='area_accumulated'):
        """
            Plot the Long Steel Bar giving the property.
        """
        y = getattr(self, prop)
        plt.plot([self.long_begin, self.long_end], [y,y])
    
    def __repr__(self):
        return str(self.__dict__)+'\n'
    
    
class LongSteelBars():
    """
        Class that defines a LongSteelBar list with easy to work properties and methods.
    """
    def __init__(self, steel_bars=[]):
        self.steel_bars = np.array(steel_bars)
        self.long_begins = np.array([ steel_bar.long_begin for steel_bar in self.steel_bars ])
        self.long_ends = np.array([ steel_bar.long_end for steel_bar in self.steel_bars ])
        self.quantities = np.array([ steel_bar.quantity for steel_bar in self.steel_bars ])
        self.diameters = np.array([ steel_bar.diameter for steel_bar in self.steel_bars ])
        self.interspaces = np.array([ steel_bar.interspace for steel_bar in self.steel_bars ])
        self.quantities_accumulated = np.array([ steel_bar.quantity_accumulated for steel_bar in self.steel_bars ])
        self.areas_accumulated = np.array([ steel_bar.area_accumulated for steel_bar in self.steel_bars ])
        self.areas = np.array([ steel_bar.area for steel_bar in self.steel_bars ])
        self.fyds = np.array([ steel_bar.fyd for steel_bar in self.steel_bars ])
        self.costs = np.array([ steel_bar.cost for steel_bar in self.steel_bars ])
        self.lengths = np.array([ steel_bar.length for steel_bar in self.steel_bars ])
        self.length = sum(self.lengths)
        self.cost = sum(self.costs)
        
    def add(self, new_steel_bars):
        """
            Add a LongSteelBar to the LongSteelBars instance.
        """
        previous_steel_bars = self.steel_bars
        if str(type(new_steel_bars)) == "<class 'fconcrete.StructuralConcrete.LongSteelBar.LongSteelBar.LongSteelBars'>":
            concatenation = list(np.concatenate((previous_steel_bars,new_steel_bars.steel_bars)))
            concatenation.sort(key=lambda x: x.long_begin, reverse=False)
            new_steel_bars = np.array(concatenation)
            
        elif str(type(new_steel_bars)) == "<class 'fconcrete.StructuralConcrete.LongSteelBar.LongSteelBar.LongSteelBar'>":
            new_steel_bars = np.append(previous_steel_bars,new_steel_bars)
        self.__init__(new_steel_bars)
    
    def changeProperty(self, prop, function, conditional=lambda x:True):
        """
            Change all properties of the LongSteelBar in a single function.
        """
        steel_bars = copy.deepcopy(self)
        for previous_steel_bar in steel_bars:
            if conditional(previous_steel_bar):
                current_attribute_value = getattr(previous_steel_bar, prop)
                setattr(previous_steel_bar, prop, function(current_attribute_value)) 
        return LongSteelBars(steel_bars.steel_bars)

    def getPositiveandNegativeLongSteelBarsInX(self, x):
        """
            Get the bars in a x longitudinal position, in cm.
            
            Returns
            -------
            positive_steel_bar_in_x : LongSteelBars
                The positive steel bar found in x.
            
            negative_steel_bar_in_x : LongSteelBars
                The negative steel bar found in x.
        """
        positive_steel_bar_in_x = np.array([steel_bar if condition else None 
                               for condition, steel_bar
                               in zip((self.long_begins<=x) & (self.long_ends>=x) & (self.areas>0), self.steel_bars)])

        negative_steel_bar_in_x = np.array([steel_bar if condition else None 
                                for condition, steel_bar
                                in zip((self.long_begins<=x) & (self.long_ends>=x) & (self.areas<0), self.steel_bars)])
        
        positive_steel_bar_in_x = LongSteelBars(positive_steel_bar_in_x[positive_steel_bar_in_x!=None])
        negative_steel_bar_in_x = LongSteelBars(negative_steel_bar_in_x[negative_steel_bar_in_x!=None])
        return positive_steel_bar_in_x, negative_steel_bar_in_x
    
    
    def getBarTransversalPosition(self, concrete_beam, x):
        """
            Get the bars in a x transversal position, in cm.
            
            Returns
            -------
            transversal_positions : array
                Each array array contains the x, y position in the transversal section, the radius of the bar and its area:
                x, y, radius, area.
        """
        _, beam_element = concrete_beam.getBeamElementInX(x)
        transversal_beam = concrete_beam.transv_steel_bars.getTransversalBarAfterX(x)
        material = beam_element.material
        section = beam_element.section
        distance_from_border = material.c+transversal_beam.diameter

        if len(self.steel_bars) == 0: return []

        transversal_positions = np.array([0,0,0,0])
        radius = max(abs(self.diameters))/2
        area = max(abs(self.areas))

        horizontal_distance = max(2, 2*radius, 1.2*concrete_beam.available_concrete.biggest_aggregate_dimension)
        vertical_distance = max(2, 2*radius, 0.5*concrete_beam.available_concrete.biggest_aggregate_dimension)

        x0, y0 = section.x0, section.y0
        x0_left_initial, x0_right_initial = x0+distance_from_border, -x0-distance_from_border
        x0_left, x0_right = x0_left_initial, x0_right_initial

        part_of_interest = (x >= self.long_begins) & (x <= self.long_ends)
        number_of_bars = max(abs(self.quantities_accumulated[part_of_interest]))
        is_positive_bar = self.areas_accumulated[part_of_interest][0] > 0

        n, bar_in_the_left, row_number = 0, True, True

        while n<number_of_bars+1:
            #y_row = y0+beam_element.material.c+radius*(row_number+1)+(vertical_distance+radius)*(row_number-1)
            y_row = y0+distance_from_border+radius+(row_number-1)*(vertical_distance+radius)
            
            if bar_in_the_left:
                x_circle, y_circle = x0_left+radius, y_row if is_positive_bar else section.height-y_row
                x0_left+=2*radius+horizontal_distance
            else:
                x_circle, y_circle = x0_right-radius, y_row if is_positive_bar else section.height-y_row
                x0_right-=2*radius+horizontal_distance
            space_between_bars = x0_right-x0_left+2*horizontal_distance
            possible_bar_in_row = (space_between_bars+horizontal_distance)//(2*radius+horizontal_distance)
            # Nao tem espaco para colcoar nenhuma depois
            if possible_bar_in_row==0:
                row_number+=1
                x0_left = x0_left_initial
                x0_right = x0_right_initial
                bar_in_the_left = False
            else: 
                plot_center = (possible_bar_in_row==1 or n == number_of_bars) and bar_in_the_left
                point = (0, y_row if is_positive_bar else section.height-y_row) if plot_center else (x_circle, y_circle)
                transversal_positions = np.vstack((transversal_positions, (point[0], point[1], radius, area)))
            n+=1
            bar_in_the_left = not bar_in_the_left

        return transversal_positions[1:]


    def plotTransversal(self, concrete_beam, x, ax=None, fig=None, color_plot="red"):
        """
            Plot the transversal vision of the longitudinal bars.
        """
        fig, ax = getAxis((-20,0), (20, 50)) if ax == None else (fig, ax)
        
        for long_bar in self.getBarTransversalPosition(concrete_beam, x):
            circle = plt.Circle((long_bar[0], long_bar[1]), long_bar[2])
            ax.add_artist(circle)
            
        return fig, ax # if return_ax else None
        
    
    def plot(self,prop='area_accumulated'):
        """
            Plot the lonfitudinal vision of the longitudinal bars.
        """
        if prop=='area_accumulated':
            plt.gca().invert_yaxis()
            
        for steelbar in self.steel_bars:
            steelbar.plot(prop)
    
    def __getitem__(self, key):
        return self.steel_bars[key]
    
    def __repr__(self):
        return str(self.steel_bars)


from fconcrete import config, duplicated, Material, Beam, Load, Node, ConcreteBeam, BeamElement, Rectangle, Concrete, Section
e = config.e
from pytest import approx
import numpy as np
import os

def approx01(x):
    return approx(x, abs=0.1)

def create_crimped_beam():
    material = Material(E=1, poisson=0.3, alpha=1)
    section = Rectangle(12,1)
    f1 = Load.PontualLoad(-1, x=500)
    n1 = Node.Crimp(x=0)
    n2 = Node.Crimp(x=1000)
    bar1 = BeamElement([n1, n2], section, material)
    return Beam(
        loads = [f1],
        beam_elements = [bar1],
        solve_structural=False
    )
    
def create_simple_beam():
    material = Material(E=1, poisson=0.3, alpha=1)
    section = Rectangle(12,1)
    f1 = Load.PontualLoad(-1, x=500)
    n1 = Node.SimpleSupport(x=0)
    n2 = Node.SimpleSupport(x=1000)
    bar1 = BeamElement([n1, n2], section, material)
    return Beam(
        loads = [f1],
        beam_elements = [bar1],
        solve_structural=False
    )
    
def test_structural_create_simple_beam():
    beam = create_simple_beam()
    assert beam.solve_structural() == None
    

def test_structural_create_crimped_beam():
    beam = create_crimped_beam()
    assert beam.solve_structural() == None

def test_structural_simple_beam():
    beam = create_simple_beam()
    beam.solve_structural()
    support_reaction = beam.nodal_efforts
    assert support_reaction[0] == approx(0.5)
    assert support_reaction[1] == approx(0)
    assert support_reaction[-2] == approx(0.5)
    assert support_reaction[-1] == approx(0)
    assert beam.getInternalMomentumStrength(beam.length/2) == approx(1*1000/4)
    assert beam.getInternalShearStrength(beam.length/2-e) == approx(0.5)
    assert beam.getInternalShearStrength(beam.length/2+e) == approx(-0.5)
    
def test_structural_crimped_beam():
    beam = create_crimped_beam()
    beam.solve_structural()
    support_reaction = beam.nodal_efforts
    assert support_reaction[0] == approx(0.5)
    assert support_reaction[1] == approx(1*1000/8)
    assert support_reaction[-2] == approx(0.5)
    assert support_reaction[-1] == approx(-1*1000/8)
    assert beam.getInternalShearStrength(beam.length/2-e) == approx(0.5)
    assert beam.getInternalShearStrength(beam.length/2+e) == approx(-0.5)
    assert beam.getInternalMomentumStrength(e) == approx(-1*1000/8, abs=1)
    assert beam.getInternalMomentumStrength(beam.length-e) == approx(-1*1000/8, abs=1)
    assert beam.getInternalMomentumStrength(beam.length/2) == approx(-1*1000/8+0.5*500, abs=1)
    
def test_structural_double_crimped_beam():
    material = Material(E=1, poisson=0.3, alpha=1)
    section = Rectangle(12,1)
    f1 = Load.PontualLoad(-200, x=200)
    n1 = Node.Crimp(x=0)
    n2 = Node.Crimp(x=1000)
    bar1 = BeamElement([n1, n2], section, material)
    beam = Beam(
        loads = [f1],
        beam_elements = [bar1]
    )
    support_reaction = beam.nodal_efforts
    assert support_reaction[0] == approx(179.2, abs=0.1)
    assert support_reaction[1] == approx(25600, abs=10)
    assert support_reaction[-2] == approx(20.8, abs=0.1)
    assert support_reaction[-1] == approx(-6400, abs=10)
    assert beam.getInternalShearStrength(150) == approx(179.2, abs=0.1)
    assert beam.getInternalShearStrength(300) == approx(-20.8, abs=0.1)
    assert beam.getInternalMomentumStrength(e) == approx(-25600, abs=10)
    assert beam.getInternalMomentumStrength(200) == approx(+10240, abs=10)
    assert beam.getInternalMomentumStrength(170) == approx(4860, abs=10)
    assert beam.getInternalMomentumStrength(210) == approx(10030, abs=10)
    assert beam.getInternalMomentumStrength(1000-e) == approx(-6400, abs=10)
    
def test_structural__crimped_simple_supported_beam():
    material = Material(E=1, poisson=0.3, alpha=1)
    section = Rectangle(12,1)
    f1 = Load.PontualLoad(-200, x=700)
    n1 = Node.Crimp(x=0)
    n2 = Node.SimpleSupport(x=1000)
    bar1 = BeamElement([n1, n2], section, material)
    beam = Beam(
        loads = [f1],
        beam_elements = [bar1]
    )
    support_reaction = beam.nodal_efforts
    assert support_reaction[0] == approx(87.3, abs=0.1)
    assert support_reaction[1] == approx(27300, abs=10)
    assert support_reaction[-2] == approx(112.7, abs=0.1)
    assert support_reaction[-1] == approx(0, abs=10)
    assert beam.getInternalShearStrength(100) == approx(87.3, abs=0.1)
    assert beam.getInternalShearStrength(800) == approx(-112.7, abs=0.1)
    assert beam.getInternalMomentumStrength(e) == approx(-27300, abs=10)
    assert beam.getInternalMomentumStrength(100) == approx(-18570, abs=10)
    assert beam.getInternalMomentumStrength(800) == approx(+22540, abs=10)
    assert beam.getInternalMomentumStrength(1000-e) == approx(0, abs=10)
    
def test_structural__crimped_simplesupported_crimped_beam():
    material = Material(E=1, poisson=0.3, alpha=1)
    section = Rectangle(12,1)
    f1 = Load.PontualLoad(-200, x=700)
    n1 = Node.Crimp(x=0)
    n2 = Node.SimpleSupport(x=1000)
    n3 = Node.Crimp(x=1500)
    bar1 = BeamElement([n1, n2], section, material)
    bar2 = BeamElement([n2, n3], section, material)
    beam = Beam(
        loads = [f1],
        beam_elements = [bar1, bar2],
    )
    support_reaction = beam.nodal_efforts
    assert support_reaction[0] == approx(57.9, abs=0.1)
    assert support_reaction[1] == approx(17500, abs=10)
    assert support_reaction[-4] == approx(200.9, abs=0.1)
    assert support_reaction[-3] == approx(0, abs=10)
    assert support_reaction[-2] == approx(-58.8, abs=0.1)
    assert support_reaction[-1] == approx(9800, abs=10)
    
    assert beam.getInternalShearStrength(100) == approx(57.9, abs=0.1)
    assert beam.getInternalShearStrength(800) == approx(-142.1, abs=0.1)
    assert beam.getInternalShearStrength(1200) == approx(58.8, abs=0.1)
    assert beam.getInternalMomentumStrength(e) == approx(-17500, abs=1)
    assert beam.getInternalMomentumStrength(100) == approx(-11710, abs=1)
    assert beam.getInternalMomentumStrength(800) == approx(8820, abs=1)
    assert beam.getInternalMomentumStrength(1200) == approx(-7840, abs=1)
    assert beam.getInternalMomentumStrength(1500-e) == approx(9800, abs=1)
    
def test_structural__crimped_simplesupported_simplesupported_beam():
    material = Material(E=1, poisson=0.3, alpha=1)
    section = Rectangle(12,1)
    f1 = Load.PontualLoad(-200, x=700)
    n1 = Node.Crimp(x=0)
    n2 = Node.SimpleSupport(x=1000)
    n3 = Node.SimpleSupport(x=1500)
    bar1 = BeamElement([n1, n2], section, material)
    bar2 = BeamElement([n2, n3], section, material)
    beam = Beam(
        loads = [f1],
        beam_elements = [bar1, bar2],
    )
    support_reaction = beam.nodal_efforts
    assert support_reaction[0] == approx(60.8, abs=0.1)
    assert support_reaction[1] == approx(18480, abs=0.1)
    assert support_reaction[-4] == approx(174.4, abs=0.1)
    assert support_reaction[-3] == approx(0, abs=0.1)
    assert support_reaction[-2] == approx(-35.30, abs=10)
    assert support_reaction[-1] == approx(0, abs=0.1)
    
    assert beam.getInternalShearStrength(100) == approx(60.8, abs=0.1)
    assert beam.getInternalShearStrength(800) == approx(-139.2, abs=0.1)
    assert beam.getInternalShearStrength(1200) == approx(35.3, abs=0.1)
    assert beam.getInternalMomentumStrength(e) == approx(-18480, abs=1)
    assert beam.getInternalMomentumStrength(100) == approx(-12400, abs=10)
    assert beam.getInternalMomentumStrength(800) == approx(10190, abs=10)
    assert beam.getInternalMomentumStrength(1200) == approx(-10580, abs=10)
    assert beam.getInternalMomentumStrength(1500-e) == approx(0, abs=10)


"""Top-level package for FConcrete."""

__author__ = """Luis Gabriel Gonçalves Coimbra"""
__email__ = 'luiscoimbraeng@outlook.com'
__version__ = '0.1.1.1'
from .helpers import *
from .Structural import *
from .StructuralConcrete import *


#!/usr/bin/env python

"""Tests for `fconcrete` package."""


import unittest
from click.testing import CliRunner

from fconcrete import fconcrete
from fconcrete import cli


class TestFconcrete(unittest.TestCase):
    """Tests for `fconcrete` package."""

    def test_structural_shear_diagram(self):
        """Test Shear Diagram"""


    def test_command_line_interface(self):
        """Test the CLI."""
        runner = CliRunner()
        result = runner.invoke(cli.main)
        assert result.exit_code == 0
        assert 'fconcrete.cli.main' in result.output
        help_result = runner.invoke(cli.main, ['--help'])
        assert help_result.exit_code == 0
        assert '--help  Show this message and exit.' in help_result.output



import numpy as np
from fconcrete.helpers import to_unit

class Load:
    """
        Class that defines a load.
    """
    def __init__(self, force, momentum, x_begin, x_end, q=0, order=0, displacement=0):
        force = to_unit(force, "kN")
        momentum = to_unit(momentum, "kNcm")
        x_begin = to_unit(x_begin, "cm")
        x_end = to_unit(x_end, "cm")
        q = to_unit(q, "kN/cm")
        
        self.x = x_begin + (x_end-x_begin)/2
        self.x_begin = x_begin
        self.x_end = x_end
        self.force = force
        self.momentum = momentum
        self.q = q
        self.order = order
        self.displacement = displacement
        
    @classmethod
    def PontualLoad(cls, load, x):
        """
            Define a pontual load.

                Call signatures:

                    fc.PontualLoad(load, x)

                >>> pontual_load_1 = fc.Load.PontualLoad(-10.0, 200)
                >>> pontual_load_2 = fc.Load.PontualLoad('-10.0kN', '2m')
                >>> repr(pontual_load_1) == repr(pontual_load_2)
                True

            Parameters
            ----------
            load : number or str
                Represent the load measure. If it is a number, default unit is kN, but also [force] unit can be given. Example:
                '20kN', '10N', etc
                
            x : number or str
                Where the load is going to end. If it is a number, default unit is cm, but also [length] unit can be given. Example:
                '20cm', '10dm', etc
            
        """ 
        return cls(load, 0, x, x, q=0, order=0)
    
    @classmethod
    def UniformDistributedLoad(cls, q, x_begin, x_end):
        """
            Define a uniform and distributed load.

                Call signatures:

                    fc.UniformDistributedLoad(q, x_begin, x_end)

                >>> uniform_load_1 = fc.Load.UniformDistributedLoad(0.1, 0, 2000)
                >>> uniform_load_2 = fc.Load.UniformDistributedLoad('10.0kN/m', '0m', '20m')
                >>> repr(uniform_load_1) == repr(uniform_load_2)
                True

            Parameters
            ----------
            q : number or str
                Represent the load by length measure. If it is a number, default unit is kN/cm, but also [force]/[length] unit can be given. Example:
                '20kN/m', '10N/m', etc
                
            x_begin : number or str
                Where the load is going to start. If it is a number, default unit is cm, but also [length] unit can be given. Example:
                '20cm', '10dm', etc
            
            x_end : number or str
                Where the load is going to end. If it is a number, default unit is cm, but also [length] unit can be given. Example:
                '20cm', '10dm', etc
        """ 
        
        q = to_unit(q, "kN/cm")
        x_begin = to_unit(x_begin, "cm")
        x_end = to_unit(x_end, "cm")
        force = q*(x_end-x_begin)
        
        return cls(force, 0, x_begin, x_end, q=q, order=1)
    
    #@classmethod
    # def DisplacementLoad(cls, x, displacement):
    #    return cls(0, 0, x, x, displacement=displacement)
    
    def __repr__(self):
        return str(self.__dict__)+'\n'


class Loads:
    """
        Class that defines a load list with easy to work properties and methods.
    """
    def __init__(self, loads):
        loads = np.array(loads)
        self.loads = loads
        self.x = np.array([ load.x for load in loads ])
        self.x_begin = np.array([ load.x_begin for load in loads ])
        self.x_end = np.array([ load.x_end for load in loads ])
        self.force = np.array([ load.force for load in loads ])
        self.momentum = np.array([ load.momentum for load in loads ])
        self.q = np.array([ load.q for load in loads ])
        self.order = np.array([ load.order for load in loads ])
    
    @classmethod
    def create(cls, loads):
        """
            Creates a instance of Loads with array of Load.
        """
        loads = np.array(loads)
        x_start = np.array([ load.x_begin for load in loads ])
        load_sort_position = np.argsort(x_start)
        return cls(loads[load_sort_position])
    
    
    def add(self, loads):
        """
            Add a array of Load in the Loads instance.
        """
        loads = np.concatenate((self.loads,loads))
        return self.create(loads)
    
    def __repr__(self):
        return str(self.loads)
     
    def __getitem__(self, key):
        return self.loads[key]
    
    def __len__(self):
        return len(self.loads)
    



from fconcrete.helpers import printProgressBar
import numpy as np
import pandas as pd

class Analysis:

    @staticmethod
    def getBestSolution(concrete_beam_function,
                max_steps_without_decrease = float("inf"),
                avoid_estimate=False,
                show_progress=True,
                sort_by_multiplication=False,
                **kwargs):
        r"""
            Returns a report with all materials and cost.
            
                Call signatures:
                        
                    `fc.Analysis.getBestSolution(concrete_beam_function,
                    ...              max_steps_without_decrease = float("inf"),
                    ...              avoid_estimate=False,
                    ...              show_progress=True,
                    ...              sort_by_multiplication=False,
                    ...              **kwargs)`
                                                            
                >>> def concrete_beam_function(width, height, length):
                ...        slab_area = 5*5
                ...        kn_per_m2 = 5
                ...        distributed_load = -slab_area*kn_per_m2/500
                ...        pp = fc.Load.UniformDistributedLoad(-width*height*25/1000000, x_begin=0, x_end=length)
                ...        n1 = fc.Node.SimpleSupport(x=0, length=20)
                ...        n2 = fc.Node.SimpleSupport(x=400, length=20)
                ...        f1 = fc.Load.UniformDistributedLoad(-0.01, x_begin=0, x_end=1)
                ...        beam = fc.ConcreteBeam(
                ...            loads = [f1, pp],
                ...            nodes = [n1, n2],
                ...            section = fc.Rectangle(width, height),
                ...            division = 200
                ...        )
                ...        return beam
                >>> full_report, solution_report, best_solution = fc.Analysis.getBestSolution(concrete_beam_function,
                ...                                     max_steps_without_decrease=15,
                ...                                     sort_by_multiplication=True,
                ...                                     avoid_estimate=True,
                ...                                     show_progress=False,
                ...                                     width=[15],
                ...                                     height=(30, 34, 2),
                ...                                     length=[150])
                >>> # Table is sorted by cost ascending, so the first one is the most economic solution.
                >>> # Alternative way to look to the best solution
                >> print(best_solution)
                {'width': 15.0, 'height': 30.0, 'length': 150.0, 'cost': 126.2650347902965, 'error': '', 'Concrete': 63.59, 'Longitudinal bar': 35.31, 'Transversal bar': 27.36}
            
            Parameters
            ----------
            concrete_beam_function
                Define the function that is going to create the beam given the parameters.
                
            max_steps_without_decrease : int, optional
                If the cost has not decrescead after max_steps_without_decrease steps, the loop breaks.
                Only use it in case your parameter combination has a logical order.
                Default inf.
            
            show_progress : `bool`, optional
                Estimate time using the last combination.
                If a exception is found, 80s per loop is set and a message about the not precision is shown.
                Also show progress bar in percentage.
                Default True.
            
            sort_by_multiplication : `bool`, optional
                Sort combinations by the multiplication os all parameter. Useful to use with max_steps_without_decrease when the is a logical order.
                Default False.
                
            kwargs
                Possible arguments for the concrete_beam_function.
                If a set of 3 elements is given, np.arange(\*kwarg_value) will be called.
                The kwargs must have the same name that the concrete_beam_function expects as arguments.
                The combination is made with np.meshgrid.
            
                
        """
        
        possible_values = []
        for kwarg_value in kwargs.values():
            possible_values = [*possible_values, np.arange(*kwarg_value)] if (len(kwarg_value) == 3 and type(kwarg_value)==tuple) else [*possible_values, kwarg_value]

        combinations = np.array(np.meshgrid(*possible_values)).T.reshape(-1,len(possible_values))

        # sorting the combinations
        if sort_by_multiplication:
            combinations_t = combinations.T
            multiply = np.ones((1,len(combinations)))
            for i in combinations_t:
                multiply = multiply*i
            combinations = np.append(combinations, multiply.T, axis=1)
            combinations = combinations[np.argsort(combinations[:,-1])][:, 0:-1]

        # combinations to dict
        combination_kwargs = [ { key: combination[i] for i, key in enumerate(kwargs.keys()) } for combination in combinations ]
        total_of_combinations = len(combination_kwargs)

        if avoid_estimate == False:
            max_variable_values = combination_kwargs[-1] 
            try:
                one_precesing_time, not_precise = concrete_beam_function(**max_variable_values).processing_time, False
            except:
                one_precesing_time, not_precise = 80, True
            continue_script = input("There are {} combinations. The estimate time to process all of them is {}s ({} minutes).{}\nType 'y' to continue or another char to cancel.\n".format(total_of_combinations, round(total_of_combinations*one_precesing_time), round(total_of_combinations*one_precesing_time/60), "\nThis measure is not precise!\n" if not_precise else ""))

        if avoid_estimate!=False or continue_script=="y":
            report = (["cost", "error", 'Concrete', 'Longitudinal bar', 'Transversal bar'])
            report = [[*list(kwargs.keys()), *report]]
            min_value, steps_without_decrease = np.inf, 0
            
            for step, combination_kwarg in enumerate(combination_kwargs):
                try:
                    beam = concrete_beam_function(**combination_kwarg)
                    error = ""
                    cost = beam.cost
                    cost_table = list(beam.subtotal_table[:, 1:2].T[0][1:].astype(float))
                    if cost != min(cost, min_value):
                        steps_without_decrease += 1
                        if steps_without_decrease >= max_steps_without_decrease:
                            if show_progress:
                                printProgressBar(total_of_combinations, total_of_combinations, prefix = 'Progress:', suffix = 'Complete', length = 50)
                                print("\n", "Ended in step {} out of {}. Because max_steps_without_decrease was reached.".format(step + 1, total_of_combinations))                        
                            break
                    else:
                        steps_without_decrease = 0
                        min_value = cost
                        
                except Exception as excep:
                    error = str(excep)
                    cost = -1
                    cost_table = [-1, -1, -1]
                
                if show_progress: printProgressBar(step + 1, total_of_combinations, prefix = 'Progress:', suffix = 'Complete', length = 50)
                row = [*combination_kwarg.values(), cost, error, *cost_table]
                report = [*report, row]
            
            full_report = pd.DataFrame(report)
            full_report.columns = full_report.loc[0]
            full_report = full_report[1:]
            
            solution_report = full_report[full_report["error"] == ""]
            solution_report = solution_report.sort_values(by="cost")
            
            best_solution = solution_report.iloc[0,:].to_dict() if len(solution_report)>0 else None
            
            return full_report, solution_report, best_solution
        


from pint import UnitRegistry
ureg = UnitRegistry()
ureg.define('kn_cm2 = kilonewton / centimeter ** 2 = kn_cm2')
ureg.define('kNcm = kilonewton * centimeter = kncm')
ureg.define('kNm = kilonewton * meter = knm')
_Q = ureg.Quantity

e = 0.00001


from fconcrete.Structural.Beam import Beam
from fconcrete.StructuralConcrete import AvailableLongConcreteSteelBar, AvailableTransvConcreteSteelBar, AvailableConcrete
from fconcrete.Structural.BeamElement import BeamElement, BeamElements
from fconcrete.helpers import timeit
import fconcrete as fc
import numpy as np
import matplotlib.pyplot as plt
import time
from fconcrete.StructuralConcrete.AvailableMaterials import solve_cost

class ConcreteBeam(Beam):
    """
        Beam associated with the material concrete.
        All attbributes from :doc:`Beam Class <../fconcrete.Structural.Beam>` can be used.
        
        Attributes
        ----------
        available_concrete : AvailableConcrete
            Same constant from input.
            Define the available concrete. 
            You can set the available fck, cost_by_m3, aggressiveness and aggregate.
            See more information in fc.AvailableConcrete docstring or the :doc:`AvailableMaterials Class <../fconcrete.StructuralConcrete.AvailableMaterials>` documentation.
            Default AvailableConcrete() which means:
            
            - 30 MPa;
            - R$353.30 by meterˆ3;
            - The aggressiveness is 3;
            - Aggregate is granite;
            - Biggest aggregate dimension is 1.5cm.
            
        available_long_steel_bars : AvailableLongConcreteSteelBar
            Same constant from input.
            Define the available longitudinal steel bars. 
            You can set the available diameters, cost_by_meter, fyw, E, etc.
            See more information in fc.AvailableLongConcreteSteelBar docstring  or the :doc:`AvailableMaterials Class <../fconcrete.StructuralConcrete.AvailableMaterials>` documentation.
            Default AvailableLongConcreteSteelBar([8]) which means:
            
            - 8mm diameter;
            - 0.5cmˆ2 area;
            - R$2.0575 by meter cost;
            - fyw equal to 50kN/cmˆ2;
            - Young Modulus (E) is 21000kN/cmˆ2;
            - Max number of steel in the section is 200;
            - Surface type is ribbed.
                
        available_transv_steel_bars : AvailableTransvConcreteSteelBar
            Same constant from input.
            Define the available transversal steel bars. 
            You can set the available diameters, cost_by_meter, fyw, E, etc.
            See more information in fc.AvailableTransvConcreteSteelBar docstring or the :doc:`AvailableMaterials Class <../fconcrete.StructuralConcrete.AvailableMaterials>` documentation.
            Default AvailableTransvConcreteSteelBar([8]) which means:
            
            - 8mm diameter;
            - 0.5cmˆ2 area;
            - R$2.0575 by meter cost;
            - The longitudinal space between transversal steel are multiple of 5;
            - fyw equal to 50kN/cmˆ2;
            - Transversal bar inclination angle of 90 degrees;
            - Tilt angle of compression struts of 45 degrees.
        
        bar_steel_max_removal : int
            Same constant from input.
            Define the max times it is possible to remove the bar.
            Default value is 100.
        
        bar_steel_removal_step : int
            Same constant from input.
            Define the step during the removal of the bar. Instead of taking the steel bars one by one, the bar_steel_removal_step will make the removal less constant.
            I makes the building process easier. 
            Default value is 2.
            
        cost : number
            Total material cost of the beam.

        cost_table : number
            Detailed table with all materials and their costs.

        design_factor : number
            Same constant from input.
            Define the number that is going to be multiplied to de momentum diagram and shear diagram.
            If your load is already a design load, you should set design_factor=1.
            Default value is 1.4.

        division : int
            Same constant from input.
            Define the number of division solutions for the beam.
            The beam will be divided in equally spaced points and all results (displacement, momentum, shear) will be calculated to these points.
            Default value is 1.4.
            
        lifetime_structure : number
            The time, in months, when the value of the deferred arrow is desired;
            Default value is 70.
                
        long_steel_bars : LongSteelBars
            Longitudinal steels used in the beam.

        long_steel_bars_solution_info : LongSteelBarSolve
            Information about the solution for longitudinal steels used in the beam.
            More information in the :doc:`LongSteelBarSolve Class <../fconcrete.StructuralConcrete.LongSteelBar.LongSteelBarSolve>` documentation.

        maximum_displacement_allowed : number
            Same constant from input.
            For each beam element, compare its maximum displacement with maximum_displacement_allowed(beam_element_length).
            This is used to solve the ELS shown in NBR 6118.
            If a beam_element length is 120cm, its maximum displacement is 1cm and maximum_displacement_allowed is 120/250=0.45cm < 1cm. Therefore, in this condition, the ELS step will raise an error.
            Default value is lambda beam_element_length : beam_element_length/250.

        processing_time : number
            Time for resolution of the concrete beam.

        subtotal_table : number
            Table with each type of material and their costs.

        tilt_angle_of_compression_struts : number
            Same constant from input.
            Tilt angle of compression struts in degrees.
            Default 45 degrees.
                
        time_begin_long_duration : number
            The time, in months, relative to the date of application of the long-term load
            Default value is 0.
                
        transv_steel_bars : TransvSteelBar
            Transversal steels used in the beam.

        transv_steel_bars_solution_info : TransvSteelBarSolve
            Information about the solution for transversal steels used in the beam.
            More information in the :doc:`TransvSteelBarSolve Class <../fconcrete.StructuralConcrete.TransvSteelBar.TransvSteelBarSolve>` documentation.

        verbose : `bool`
            Print the the steps and their durations.
            Default value is False.
    """
    def __init__(self,
                 loads,
                 beam_elements=None,
                 nodes=None,
                 section=None,
                 design_factor=1.4,
                 division=1000,
                 maximum_displacement_allowed=lambda beam_element_length : beam_element_length/250,
                 available_long_steel_bars=AvailableLongConcreteSteelBar([8]),
                 bar_steel_removal_step=2,
                 bar_steel_max_removal=100,
                 available_transv_steel_bars=AvailableTransvConcreteSteelBar([8]),
                 tilt_angle_of_compression_struts=45,
                 available_concrete=AvailableConcrete(),
                 time_begin_long_duration=0,
                 lifetime_structure=70,
                 verbose = False,
                 max_relative_diff_of_steel_height = 0.02,
                 **options):
        """
            Returns a concrete_beam element.
            
                Call signatures:

                    `ConcreteBeam(loads,
                                beam_elements=None,
                                nodes=None,
                                section=None,
                                bar_steel_removal_step=2,
                                bar_steel_max_removal=100,
                                design_factor=1.4,
                                division=1000,
                                maximum_displacement_allowed=lambda beam_element_length : beam_element_length/250,
                                tilt_angle_of_compression_struts=45,
                                available_long_steel_bars=AvailableLongConcreteSteelBar(),
                                available_transv_steel_bars=AvailableTransvConcreteSteelBar(),
                                available_concrete=AvailableConcrete(),
                                time_begin_long_duration=0,
                                lifetime_structure=70,
                                verbose = False,
                                **options)`

                >>> n1 = fc.Node.SimpleSupport(x=0, length=20)
                >>> n2 = fc.Node.SimpleSupport(x=400, length=20)
                >>> f1 = fc.Load.UniformDistributedLoad(-0.000001, x_begin=0, x_end=1)
                 
                >>> concrete_beam = fc.ConcreteBeam(
                >>>     loads = [f1],
                >>>     nodes = [n1, n2],
                >>>     section = fc.Rectangle(20,1000),
                >>>     division = 20
                >>> )
            
            Parameters
            ----------
            loads : [Load]
                Define the loads supported for the beam.
            
            beam_elements : [BeamElement], optional
                Define the beam_elements that, together, makes the whole Beam. 
                Optional if nodes and section is given.
            
            nodes : [Node]
                Define the nodes that are going to make the whole Beam.
                Not used if beam_elements is given.
                
            section : Section
                Define the section that are going to make the whole Beam.
                Not used if beam_elements is given.
            
            design_factor : number, optional
                Define the number that is going to be multiplied to de momentum diagram and shear diagram.
                If your load is already a design load, you should set design_factor=1.
                Default value is 1.4.
                
            division : int, optional
                Define the number of division solutions for the beam.
                The beam will be divided in equally spaced points and all results (displacement, momentum, shear) will be calculated to these points.
                Default value is 1.4.
            
            maximum_displacement_allowed : number, optional
                For each beam element, compare its maximum displacement with maximum_displacement_allowed(beam_element_length).
                This is used to solve the ELS shown in NBR 6118.
                If a beam_element length is 120cm, its maximum displacement is 1cm and maximum_displacement_allowed is 120/250=0.45cm < 1cm. Therefore, in this condition, the ELS step will raise an error.
                Default value is lambda beam_element_length : beam_element_length/250.
                
            available_long_steel_bars : AvailableLongConcreteSteelBar, optional
                Define the available longitudinal steel bars. 
                You can set the available diameters, cost_by_meter, fyw, E, etc.
                See more information in fc.AvailableLongConcreteSteelBar docstring  or the :doc:`AvailableMaterials Class <../fconcrete.StructuralConcrete.AvailableMaterials>` documentation.
                Default AvailableLongConcreteSteelBar([8]) which means:
                
                - 8mm diameter;
                - 0.5cmˆ2 area;
                - R$2.0575 by meter cost;
                - fyw equal to 50kN/cmˆ2;
                - Young Modulus (E) is 21000kN/cmˆ2;
                - Max number of steel in the section is 200;
                - Surface type is ribbed.
                
            bar_steel_removal_step : int, optional
                Define the step during the removal of the bar. Instead of taking the steel bars one by one, the bar_steel_removal_step will make the removal less constant.
                I makes the building process easier. 
                Default value is 2.
                
            bar_steel_max_removal : int, optional
                Define the max times it is possible to remove the bar.
                Default value is 100.
                
            available_transv_steel_bars : AvailableLongConcreteSteelBar
                Define the available longitudinal steel bars. 
                You can set the available diameters, cost_by_meter, fyw, E, etc.
                See more information in fc.AvailableLongConcreteSteelBar docstring or the :doc:`AvailableMaterials Class <../fconcrete.StructuralConcrete.AvailableMaterials>` documentation.
                Default AvailableLongConcreteSteelBar([8]) which means:
                
                - 8mm diameter;
                - 0.5cmˆ2 area;
                - R$2.0575 by meter cost;
                - The longitudinal spaces between transversal steel are multiple of 5;
                - fyw equal to 50kN/cmˆ2;
                - Transversal bar inclination angle of 90 degrees;
                - Tilt angle of compression struts of 45 degree.
            
            tilt_angle_of_compression_struts : number
                Tilt angle of compression struts in degrees.
                Default 45 degrees.
            
            available_concrete : AvailableConcrete
                Define the available concrete. 
                You can set the available fck, cost_by_m3, aggressiveness and aggregate.
                See more information in fc.AvailableConcrete docstring or the :doc:`AvailableMaterials Class <../fconcrete.StructuralConcrete.AvailableMaterials>` documentation.
                Default AvailableConcrete() which means:
                
                - 30 MPa;
                - R$353.30 by meterˆ3;
                - The aggressiveness is 3;
                - Aggregate is granite.
                - Biggest aggregate dimension is 1.5cm.
            
            time_begin_long_duration : number, optional
                The time, in months, relative to the date of application of the long-term load
                Default value is 0.
            
            lifetime_structure : number, optional
                The time, in months, when the value of the deferred arrow is desired;
                Default value is 70.
            
            verbose : bool, optional
                Print the the steps and their durations.
                Default value is False.
            
            max_relative_diff_of_steel_height: number, optional
                Maximum value for relative difference of the beam section "d" value.
                The relative difference is calculated taking the module of the sum of all previous d's less the sum for the calculated value of d divided by the sum of all previous calculated d's
                If this values is greater than the max_relative_diff_of_d, all concrete_beam is recalculated.
                The initial value of d is set to be 0.8*height.
                Default value is 0.02.
        """
        start = time.time()
        
        beam_elements = self._checkInput(
            nodes=nodes,
            beam_elements=beam_elements,
            material=available_concrete.material,
            section=section
        )
        
        timeit(verbose, "Solve structural beam")(Beam.__init__(self, loads, beam_elements, solve_displacement=False, **options))
        
        self.bar_steel_removal_step = bar_steel_removal_step
        self.bar_steel_max_removal = bar_steel_max_removal
        self.design_factor = design_factor
        self.division = division
        self.maximum_displacement_allowed = maximum_displacement_allowed
        self.tilt_angle_of_compression_struts = tilt_angle_of_compression_struts
        self.available_long_steel_bars = available_long_steel_bars
        self.available_transv_steel_bars = available_transv_steel_bars
        self.available_concrete = available_concrete
        self.time_begin_long_duration = time_begin_long_duration
        self.lifetime_structure = lifetime_structure
        self.verbose = verbose
        self.max_relative_diff_of_steel_height = max_relative_diff_of_steel_height
            
        if options.get("solve_transv_steel") != False:
            timeit(verbose, "Solve transv steel")(self.solve_transv_steel)()
            
        if options.get("solve_long_steel") != False:
            self.solve_long_steel()
        
        if options.get("solve_ELS") != False:
            timeit(verbose, "Solve ELS")(self.solve_ELS)()
        
        if options.get("solve_cost") != False:
            self.solve_cost()
            # d value is the initially with 0.8*height. This function check if initial guess is ok.
            self.checkRecalculationOfD()
        
        end = time.time()
        self.processing_time = end-start
    
    def getConcreteDisplacementDiagram(self, **options):
        """
            Returns necessary steel area given the position and momentum.
            
            Parameters
            ----------
            **options
                
                ``division``:
                    Number of divisions equally spaced (`int`).
                ``x_begin``:
                    Begin of the x_axis (`number`).
                ``x_end``:
                    End of the x_axis (`number`).
                
            Returns
            -------
            x : list of number
                X axis in cm.
                
            displacement : list of number
                Vertical displacement value in cm.
                
        """ 
        x, y = self.getDisplacementDiagram(**options)
        return x, y*(self._time_function_coefficient(self.lifetime_structure)-self._time_function_coefficient(self.time_begin_long_duration))
    
    def plotConcreteDisplacementDiagram(self, **options):
        """
            Apply concrete_beam.getConcreteDisplacementDiagram for options["division"] parts of the beam.
            
            Parameters
            ----------
            **options
                
                ``division``:
                    Number of divisions equally spaced (`int`).
                ``x_begin``:
                    Begin of the x_axis (`number`).
                ``x_end``:
                    End of the x_axis (`number`).
                
            Returns
            -------
            x : list of number
                The x position of the division in cm
            
            y : list of number
                The value of displacement for each x.
        """
        options["division"] = options["division"] if options.get("division") else self.division
        x, y = self.getConcreteDisplacementDiagram(**options)
        plt.plot(x, y)
            
    @staticmethod
    def _time_function_coefficient(t):
        if t>70: return 2
        return 0.68*(0.996**t)*t**0.32 
    
    def solve_ELS(self):
        """
            Starts the process of solution for ELS (Estado Limite de Serviço)
        """
        self.initial_beam_elements = self._toConcreteBeamElements(self.initial_beam_elements)
        self.solve_displacement()
        for beam_element in self.initial_beam_elements:
            x_begin = beam_element.n1.x
            x_end = beam_element.n2.x
            _, y = self.getConcreteDisplacementDiagram(x_begin=x_begin, x_end=x_end, division=200)
            max_disp = self.maximum_displacement_allowed(beam_element.length)
            if max(abs(y)) > max_disp:
                raise Exception("Displacement too big between x={}cm and x={}cm. Maximum allowed is {}cm, but the beam lement reached {}cm".format(
                    x_begin, x_end, max_disp, max(abs(y))))
    
    def getShearDesignDiagram(self, **options):
        """
            Apply beam.getShearDiagram for options["division"] parts of the beam and multiplies by concrete_beam.design_factor.
            
            Parameters
            ----------
            **options
                
                ``division``:
                    Number of divisions equally spaced (`int`). Default concrete_beam.division.
                ``x_begin``:
                    Begin of the x_axis (`number`).
                ``x_end``:
                    End of the x_axis (`number`).
                
                
            Returns
            -------
            x : list of number
                The x position of the division in cm
            
            y : list of number
                The value of shear for each x.
        """
        options["division"] = options["division"] if options.get("division") else self.division
        x, shear_diagram = self.getShearDiagram(**options)
        return x, self.design_factor*shear_diagram
    
    def plotShearDesignDiagram(self, **options):
        """
            Simply applies the beam.getShearDesignDiagram method results (x,y) to a plot with plt.plot(x, y).

            Parameters
            ----------
            **options
                
                ``division``:
                    Number of divisions equally spaced (`int`).
                ``x_begin``:
                    Begin of the x_axis (`number`).
                ``x_end``:
                    End of the x_axis (`number`).
                
        """
        x, y = self.getShearDesignDiagram(**options)
        plt.plot(x, y)
    
    def plotTransversalInX(self, x):
        """
            Plot an image of the transversal section with the longitudinal and transversal steel.

                Call signatures:

                    concrete_beam.plotTransversalInX.getSteelArea(x)
                    
            Returns
            -------
            fig
                Figure generated by matplotlib.
                
            ax
                Axis generated by matplotlib.
                
        """
        positive_bars, negative_bars = self.long_steel_bars.getPositiveandNegativeLongSteelBarsInX(x=x)
        transversal_bar = self.transv_steel_bars.getTransversalBarAfterX(x)
        
        _, beam_element = self.getBeamElementInX(x)
        material, section = beam_element.material, beam_element.section
        
        fig, ax = section.plot()
        fig, ax = transversal_bar.plot(fig=fig, ax=ax, c=material.c)
        fig, ax = positive_bars.plotTransversal(self, x, fig=fig, ax=ax)
        fig, ax = negative_bars.plotTransversal(self, x, fig=fig, ax=ax)
        return fig, ax

    def solve_transv_steel(self):
        """
            Starts the process of solution for the used transversal steel.
        """
        self.transv_steel_bars_solution_info = fc.TransvSteelBarSolve(concrete_beam=self,
                                                                        fyk=self.available_transv_steel_bars.fyw,
                                                                        theta_in_degree= self.tilt_angle_of_compression_struts,
                                                                        alpha_in_degree = self.available_transv_steel_bars.inclination_angle)
        self.transv_steel_bars = self.transv_steel_bars_solution_info.steel_bars
    
    def solve_long_steel(self):
        """
            Starts the process of solution for the used longitudinal steel.
        """
        self.long_steel_bars_solution_info = fc.LongSteelBarSolve(concrete_beam=self)
        self.long_steel_bars = self.long_steel_bars_solution_info.steel_bars
    
    @staticmethod
    def _checkInput(**inputs):
        nodes, beam_elements, section, material = inputs.get("nodes"), inputs.get("beam_elements"), inputs.get("section"), inputs.get("material")
        if nodes and section:
            section = fc.ConcreteSection.setSteelHeight(section)
            if len(nodes) == 1: raise Exception("Must contain at least 2 nodes to create a beam")
            beam_elements = []
            for i in range(0,len(nodes)-1):
                beam_elements = [*beam_elements, fc.BeamElement([nodes[i], nodes[i+1]], section, material)]
        elif beam_elements and section:
            beam_elements = BeamElements.create(beam_elements)
            beam_elements = beam_elements.changeProperty("material", lambda x:material)
            section = fc.ConcreteSection.setSteelHeight(section)
            beam_elements = beam_elements.changeProperty("section", lambda x:section)
        elif beam_elements:
            beam_elements_modified = []
            for beam_element in beam_elements:
                #.positive_steel_height, beam_element.section.negative_steel_height, beam_element.section.maximum_steel_height, beam_element.section.minimum_steel_height  = 0.8*beam_element.section.height, 0.8*beam_element.section.height, 0.8*beam_element.section.height, 0.8*beam_element.section.height
                section = fc.ConcreteSection.setSteelHeight(beam_element.section)
                beam_element.section = section
                beam_elements_modified = [*beam_elements_modified, beam_element]
            beam_elements = beam_elements_modified
        
        return beam_elements
    
    def _toConcreteBeamElements(self, beam_elements):
        for beam_element in beam_elements:
            x_begin = beam_element.n1.x
            x_end = beam_element.n2.x
            section = beam_element.section
            material = beam_element.material
            
            I = section.I
            y_cg = section.y_cg
            bw = section.bw
            d = section.positive_steel_height
            
            h = section.height
            
            fctm = material.fctm
            E_cs = material.E_cs
            E_s = self.available_long_steel_bars.E
            is_in_the_beam_element = (self.long_steel_bars_solution_info.x >= x_begin) &  (self.long_steel_bars_solution_info.x <= x_end)
            positive_area_info = self.long_steel_bars_solution_info.positive_areas_info[:, is_in_the_beam_element]
            negative_area_info = self.long_steel_bars_solution_info.negative_areas_info[:, is_in_the_beam_element]

            positive_area_diagram = positive_area_info[2]
            negative_area_diagram = negative_area_info[2]
            positive_area_diagram = positive_area_diagram[~np.isnan(positive_area_diagram)]
            negative_area_diagram = negative_area_diagram[~np.isnan(negative_area_diagram)]
            
            max_positive_area = abs(positive_area_diagram).max(initial=0)
            max_negative_area = abs(negative_area_diagram).max(initial=0)
            max_area = max(max_positive_area, max_negative_area)
            
            # fator que correlaciona aproximadamente a resistência à tração na flexão com a resistência à tração direta
            y_t = y_cg if max_area==max_positive_area else h-y_cg
            alpha = 1.5
            M_r = alpha*fctm*I/y_t
            _, momentum_diagram = self.getMomentumDiagram(x_begin=x_begin, x_end=x_end, division=200)
            M_a = self.design_factor*max(abs(momentum_diagram))
            
            mra3 = (M_r/M_a)**3
            alpha_e = E_s/E_cs
            
            a1, a2, a3 = bw/2, max_area*alpha_e, -max_area*alpha_e*d
            x2 = (-a2+(a2**2-4*a1*a3)**0.5)/(2*a1)
            I2 = bw*x2**3/3+a2*(x2-d)**2
            
            new_I = min((mra3*I+(1-mra3)*I2),I)
            beam_element.I = new_I
            beam_element.flexural_rigidity = E_cs * new_I
            
        return beam_elements
    
    def solve_cost(self):
        self.cost, self.cost_table, self.subtotal_table = solve_cost(self)
        
    def checkRecalculationOfD(self):
        while True:
            x_changes = np.concatenate((self.long_steel_bars.long_begins, self.long_steel_bars.long_ends))
            x_changes = x_changes[np.isin(x_changes, self.beam_elements.nodes.x, invert=True)]
            x_changes = np.unique(x_changes[(x_changes>=0) & (x_changes<=self.length)])

            nodes_change = [ fc.Node.Crimp(x) for x in x_changes ]
            new_nodes = fc.Nodes(np.concatenate((nodes_change, self.beam_elements.nodes)))
            new_nodes = new_nodes[np.argsort(new_nodes.x)]

            previous_ds_positive, previous_ds_negative, diff_positive, diff_negative = 0, 0, 0, 0
            beam_elements = []

            for i in range(0, len(new_nodes)-1):
                current_x = new_nodes[i].x
                next_x = new_nodes[i+1].x
                middle_x = (current_x+next_x)/2

                positive_bars, negative_bars = self.long_steel_bars.getPositiveandNegativeLongSteelBarsInX(x=middle_x)
                positive_transversal_position, negative_transversal_position = positive_bars.getBarTransversalPosition(self, x=middle_x), negative_bars.getBarTransversalPosition(self, x=middle_x)

                transversal_position = negative_transversal_position if len(positive_transversal_position)==0 else (
                    positive_transversal_position if len(negative_transversal_position)==0 else np.concatenate((
                        positive_transversal_position,
                        negative_transversal_position
                    ))
                )

                _, beam_element = self.getBeamElementInX(middle_x)
                section, material = beam_element.section, beam_element.material
                height = section.height

                _, y, _, area = transversal_position.T

                y_c_negative = (y[y >= (height/2)] @ area[y >= (height/2)])/sum(area[y >= (height/2)]) if sum(y >= (height/2)) else 0
                y_c_positive = height - (y[y < (height/2)] @ area[y < (height/2)])/sum(area[y < (height/2)]) if sum(y < (height/2)) else 0

                new_positive_steel_height, previous_positive_steel_height = (y_c_positive), section.positive_steel_height
                new_negative_steel_height, previous_negative_steel_height = (y_c_negative), section.negative_steel_height

                previous_ds_positive += 0 if new_positive_steel_height else previous_positive_steel_height
                previous_ds_negative += 0 if new_negative_steel_height else previous_negative_steel_height
                diff_positive += abs(previous_positive_steel_height-new_positive_steel_height if new_positive_steel_height else 0)
                diff_negative += abs(previous_negative_steel_height-new_negative_steel_height if new_negative_steel_height else 0)

                new_section = fc.Rectangle(section.width(), section.height)
                new_section = fc.ConcreteSection.setSteelHeight(new_section, new_positive_steel_height, new_negative_steel_height)

                new_beam_element = fc.BeamElement([new_nodes[i], new_nodes[i+1]], new_section, material)
                beam_elements = [*beam_elements, new_beam_element]
            
            relative_positive_diff = diff_positive/previous_ds_positive if previous_ds_positive else 0
            relative_negative_diff = diff_negative/previous_ds_negative if previous_ds_negative else 0
            
            if max(relative_positive_diff, relative_negative_diff) > self.max_relative_diff_of_steel_height:
                self.beam_elements = fc.BeamElements(beam_elements)
                self.solve_long_steel()
                self.solve_transv_steel()
                self.solve_ELS()
                self.solve_cost()
            else:
                break
            
    def __name__(self):
        return "ConcreteBeam"


from fconcrete.helpers import cond, integrate, to_unit
import numpy as np
import matplotlib.pyplot as plt
from fconcrete.helpers import getAxis

class Section():
    """
        Class to represent simetrical section along the y axis.
        
        Attributes
        ----------
        height : number
            Maximum height of the section in cm.
            
        function_width : function
            Define the width along the y axis. The function starts with x=0 and ends in x=height.
         
        area : number
            Total area of the section in cmˆ2.
        
        I : number
            Moment of inertia in cmˆ4.
            
        x0 : number
            Initial reference in the x axis.
            
        y0 : number
            Initial reference in the y axis.
    """    
    def __init__(self, function_width, height):
        """
            Creates simetrical section along the y axis.
            
            Parameters
            ----------
            function_width : function
                Define the width along the y axis. The function starts with x=0 and ends in x=height.
                
            height : number
                Represent the maximum y value possible in cm.
        """ 
        height = to_unit(height, "cm")
        self.height = height
        self.function_width = function_width
        self.area = self.getAreaBetween(0, height, 1000)
        self.I = self._I()
        self.x0, self.y0 = -self.function_width(0), 0
        #self.bw = 2*min(abs(function_width))
        
    def width(self, y):
        """
            Gets the width in y.
        """
        return 0 if y>self.height else self.function_width(y)
    
    def getAreaBetween(self, begin_height, end_height, interations=100):
        """
            Area between 2 y values.
        """
        return 2*integrate(self.function_width, begin_height, end_height, interations)
    
    def _I(self):
        raise NotImplementedError
        
    def plot(self, N=100, color_plot="red", ax=None, fig=None):
        """
            Plot the section.
        """
        height = self.height
        
        fig, ax = getAxis() if ax == None else (fig, ax)
        y = np.linspace(0, height, N)
        x = np.array([self.function_width(xi) for xi in y])

        x_base = np.linspace(-x[0], x[0], 100)
        x_top = np.linspace(-x[-1], x[-1], 100)

        ax.plot(x, y, color=color_plot) # plot right simetry
        ax.plot(2*y[0]-x, y, color=color_plot) # plot left simetry
        ax.plot(x_base,np.zeros(len(y)), color=color_plot) # plot base
        ax.plot(x_top,height*np.ones(len(y)), color=color_plot) # plot top
        
        return fig, ax
    
    
class Rectangle(Section):
    """
        Attributes
        ----------
        height : number
            Maximum height of the section in cm.
            
        function_width : function
            Define the width along the y axis. The function starts with x=0 and ends in x=height.
         
        bw : number
            Minimum width in cm.
        
        area : number
            Total area of the section in cmˆ2.
        
        I : number
            Moment of inertia in cmˆ4.
        
        y_cg : number
            Gravity center in the y axis.
        
        x0 : number
            Initial reference in the x axis.
            
        y0 : number
            Initial reference in the y axis.
    """
    def __init__(self,width, height):
        """
            Returns a concrete_beam element.
            
                Call signatures:

                    fc.Rectangle(width, height)

                >>> section = fc.Rectangle(25,56)
                
            Parameters
            ----------
            width : number
            height : number
        """    
        width = to_unit(width, "cm")
        height = to_unit(height, "cm")
        self.__width = width
        self.height = height
        self.function_width = lambda x:width/2
        self.area = width*height
        self.I = self.width()*self.height**3/12
        self.bw = width
        self.y_cg = height/2
        self.x0 = -width/2
        self.y0 = 0
        
    def getAreaBetween(self, begin_height, end_height):
        """
            Area between 2 y values.
        """
        return self.width()*(end_height - begin_height)
    
    def width(self, height=0):
        """
            Width value in cm.
        """
        return self.__width

    def __name__(self):
        return "Rectangle"
    
    
unitary_section = Rectangle(12, 10)
    


import numpy as np
from .TransvSteelBar import TransvSteelBar, TransvSteelBars
from math import radians, sin, tan

class TransvSteelBarSolve():
    def __init__(self, concrete_beam, fyk=50, theta_in_degree=45, alpha_in_degree = 90):
        self.available = concrete_beam.available_transv_steel_bars
        self.concrete_beam = concrete_beam
        x, shear_diagram = concrete_beam.getShearDesignDiagram(division=concrete_beam.division)
        self.x = x
        self.shear_diagram = abs(shear_diagram)
        self.fyk = fyk
        self.fyd = fyk/1.4
        self.theta = radians(theta_in_degree)
        self.alpha = radians(alpha_in_degree)
        v_rd2, d, v_sd = self.checkProbableCompressedConnectingRod()
        self.s_max = self._getS_max(v_rd2, d, v_sd, self.available)
        _, self.shear_area_per_cm = self.getShearSteelAreaPerCmDiagram()
        self.steel_bars = self.getStirrupsInfo()
    
    @staticmethod
    def _getS_max(v_rd2, d, v_sd, available):
        s_max = min(0.6*d, 30) if v_sd <= 0.67*v_rd2 else min(0.3*d, 20)        
        return s_max
        
    def checkProbableCompressedConnectingRod(self):
        """
            Check probable compressed connecting rod.
            It is probable because checks only where the shear is maximum.

                Call signatures:

                    concrete_beam.transv_steel_bars_solution_info.checkProbableCompressedConnectingRod()

            Returns
            -------
            v_rd2 : number
                Shear of calculation, related to the ruin of compressed concrete diagonals in kN.
                
            d : number
                Distance from longitudinal steel bars to the other extremity of the section in cm.
            
            max_shear : number
                Maximum shear in kN.
        """
        max_shear = max(self.shear_diagram)
        max_shear_x = self.x[self.shear_diagram == max_shear][0]
        _, single_beam_element = self.concrete_beam.getBeamElementInX(max_shear_x)
        v_rd2 = self.getV_rd2(single_beam_element)
        check = max_shear <= v_rd2
        if check == False: raise Exception("Shear ({}kN) in x={} is greater or equal to maximum shear allowed ({}kN)".format(max_shear, max_shear_x, v_rd2))
        d = single_beam_element.section.minimum_steel_height
        
        return v_rd2, d, max_shear
    
    def getV_rd2(self, single_beam_element):
        """
            Giving a beam element, calculates the shear related to the ruin of compressed concrete diagonals in kN.
        """
        fck = single_beam_element.material.fck
        bw = single_beam_element.section.bw
        d = single_beam_element.section.minimum_steel_height
        
        fcd = single_beam_element.material.fcd
        alpha_v2 = (1-fck/25)
        v_rd2 = 0.54*alpha_v2*fcd*bw*d*(sin(self.theta))*(tan(self.alpha)**(-1)+tan(self.theta)**(-1))
        return v_rd2
    
    def getMinimumSteelAreaPerCm(self,single_beam_element):
        """
            Giving a beam element, calculates the minimum steel area (cmˆ2) per cm.
        """
        fctm = single_beam_element.material.fctm
        bw = single_beam_element.section.bw
        As_per_cm_min = 0.2*fctm*bw*sin(self.alpha)/self.fyk
        return As_per_cm_min
    

    def getShearSteelAreaPerCm(self, x, v_sd): 
        """
            Calculates the shear steel area (cmˆ2) per cm considering the restrictions.
        """
        # can be optimized
        _, single_beam_element = self.concrete_beam.getBeamElementInX(x)
        v_rd2 = self.getV_rd2(single_beam_element)
        As_per_cm_min = self.getMinimumSteelAreaPerCm(single_beam_element)
        
        _, single_beam_element = self.concrete_beam.getBeamElementInX(x)
        bw = single_beam_element.section.bw
        d = single_beam_element.section.minimum_steel_height
        fctd = single_beam_element.material.fctd

        v_c0 = 0.6*fctd*bw*d
        v_c1 = np.interp(v_sd, [v_c0, v_c0, v_rd2], [v_c0, v_c0, 0])
        v_c = v_c1
        v_sw = max(v_sd - v_c, 0)
        
        As_per_cm = v_sw/(0.9*d*self.fyd*(sin(self.alpha))*(tan(self.alpha)**(-1)+tan(self.theta)**(-1)))
        
        return max(As_per_cm, As_per_cm_min)

    def getShearSteelAreaPerCmDiagram(self):
        """
            Apply concrete_beam.transv_steel_bars_solution_info.getShearSteelAreaPerCm for parts of the concrete_beam.
            
            Returns
            -------
            x : list of number
                The x position of the division in cm
            
            y : list of number
                The value of shear area per cm for each x.
        """
        shear_area_per_cm = [self.getShearSteelAreaPerCm(x_u, v_sd) for x_u, v_sd in zip(self.x, self.shear_diagram)]
        return self.x, np.array(shear_area_per_cm)
    
    def getComercialInfo(self, as_per_cm):
        """
            Get comercial info giving the area per cm.
            
            Returns
            -------
            diameter : number
                Diameter in cm.
            
            space : number
                Longitudinal space between the transversal steel.
                
            area : number
                Area of the transversal steel bar in cmˆ2.
                
            as_per_cm : number
                Area of the transversal steel bar in cmˆ2 per cm.
        """
        table = self.available.table
        comercial_info = table[table[:, 3] >= as_per_cm]
        if len(comercial_info) == 0 : raise Exception("It is not possible to place the transversal steel bar using the provided space_in_multiple_of or diameter argument. When you create the concrete_beam, you should change the argument 'available_transv_steel_bars = fc.AvailableTransvConcreteSteelBar(diameters=[x], space_is_multiple_of=[y])' giving y a smaller number or diameter a bigger one.")
        diameter, space, area, as_per_cm = comercial_info[0]
        return diameter, space, area, as_per_cm
    
    def getStirrupsInfo(self):
        """
            Format all informations and return a TransvSteelBars instance.
        """
        x_array, shear_area_per_cm = self.x, self.shear_area_per_cm
        s_max = self.s_max
        x = self.concrete_beam.x_begin
        transversal_steel = TransvSteelBars()

        while x<self.concrete_beam.x_end:
            x_loop = ((x_array>x) & (x_array<x+s_max))
            shear = max(abs(shear_area_per_cm[x_loop]))
            diameter, space, area, as_per_cm = self.getComercialInfo(shear)
            single_beam_element = self.concrete_beam.getBeamElementInX(x)[1]
            height = single_beam_element.section.height
            width = single_beam_element.section.width()
            c = single_beam_element.material.c
            anchor = max(5*diameter, 5)
            length = width*2+height*2+anchor
            cost = length*self.available.cost_by_meter[diameter*10]/100
            
            transversal_steel.add(
                TransvSteelBar(
                    x=x,
                    height=height-2*c,
                    width=width-2*c,
                    diameter=diameter,
                    space_after=space,
                    area=area,
                    as_per_cm=as_per_cm,
                    anchor = anchor,
                    length=length,
                    cost=cost)
            )
            x += space
        
        # Add by the end of the beam
        transversal_steel.add(
                TransvSteelBar(
                    x=self.concrete_beam.x_end,
                    height=height-2*c,
                    width=width-2*c,
                    diameter=diameter,
                    space_after=space,
                    area=area,
                    as_per_cm=as_per_cm,
                    anchor = anchor,
                    length=length,
                    cost=cost)
        )
        
        return transversal_steel


import numpy as np
from .LongSteelBar import LongSteelBar, LongSteelBars
import warnings
from scipy.signal import find_peaks
from math import radians, sin, tan
from fconcrete.helpers import timeit

class LongSteelBarSolve():
    def __init__(self, concrete_beam):
        verbose = concrete_beam.verbose
        self.verbose = verbose
        self.available = concrete_beam.available_long_steel_bars
        self.concrete_beam = concrete_beam
        
        self.bar_steel_removal_step = self.concrete_beam.bar_steel_removal_step
        self.bar_steel_max_removal = self.concrete_beam.bar_steel_max_removal
        
        x, positive_areas_info, negative_areas_info = timeit(verbose)(self.getComercialSteelAreaDiagram)(division=concrete_beam.division)
        self.x = x
        self.positive_areas_info = positive_areas_info
        self.negative_areas_info = negative_areas_info
        
        interspace_between_momentum_positive = self._getInterspaceBetweenMomentum(x, area=positive_areas_info[2])
        interspace_between_momentum_negative = self._getInterspaceBetweenMomentum(x, area=negative_areas_info[2])
        
        self.interspace_between_momentum_positive = interspace_between_momentum_positive
        self.interspace_between_momentum_negative = interspace_between_momentum_negative
        
        steel_bars_positive = self._getBarsInInterspaces(x, positive_areas_info, interspace_between_momentum_positive)
        steel_bars_negative = self._getBarsInInterspaces(x, negative_areas_info, interspace_between_momentum_negative)
        #steel_bars_negative = self._getBarsInInterspaces(x, negative_areas_info, interspace_between_momentum_negative)
        
        concatenation = list(np.concatenate((steel_bars_positive.steel_bars,steel_bars_negative.steel_bars)))
        concatenation.sort(key=lambda x: x.long_begin, reverse=False)
        steel_bars = LongSteelBars(np.array(concatenation))
        
        steel_bars_with_anchor_length_positive = timeit(verbose, "Anchor positive steel bar")(self._anchorSteelBars)(steel_bars, interspace_between_momentum_positive)
        self.steel_bars = timeit(verbose, "Anchor negative steel bar")(self._anchorSteelBars)(steel_bars_with_anchor_length_positive, interspace_between_momentum_negative)
        
        
    def getDecalagedMomentumDesignDiagram(self, **options_diagram):
        """
            Returns tuple with 3 np.array: x (axis), momentum_positive, momentum_negative.
            
                Call signatures:

                    concrete_beam.long_steel_bars_solution_info.getDecalagedMomentumDesignDiagram(division=1000)

                >>> x_decalaged, momentum_positive, momentum_negative = concrete_beam.long_steel_bars_solution_info.getDecalagedMomentumDesignDiagram(division=100)
            
            Parameters
            ----------
            division : int, optional (default 1000)
                Define the step to plot the graph.
                A high number means a more precise graph, but also you need more processing time.
            
        """
        x, momentum_diagram = self.concrete_beam.getMomentumDiagram(division=self.concrete_beam.division)
        x_decalaged, decalaged_x_left, decalaged_x_right, join_decalaged_x_order = self.__decalageds_x_axis(x)
        momentum_positive, momentum_negative = self.__decalaged_momentums(x_decalaged,
                                                                    decalaged_x_left,
                                                                    decalaged_x_right,
                                                                    join_decalaged_x_order,
                                                                    momentum_diagram)
        momentum_positive = self.__join_momentum_peak(momentum_positive)
        momentum_negative = self.__join_momentum_peak(momentum_negative)
        
        return x_decalaged, self.concrete_beam.design_factor*momentum_positive, self.concrete_beam.design_factor*momentum_negative
    
    
    def __decalageds_x_axis(self, x):
        decalaged_x_left = np.array([])
        decalaged_x_right = np.array([])
        
        for beam_element in self.concrete_beam.beam_elements:
            a_l = self.getDecalagedLength(beam_element)
            position_in_beam_element = x[(x>=beam_element.n1.x) & (x<=beam_element.n2.x)]
            decalaged_x_left_temp = position_in_beam_element - a_l
            decalaged_x_right_temp = position_in_beam_element + a_l
            decalaged_x_left = np.concatenate((decalaged_x_left, decalaged_x_left_temp))
            decalaged_x_right = np.concatenate((decalaged_x_right, decalaged_x_right_temp))
        
        join_decalaged_x = np.concatenate((decalaged_x_left, x, decalaged_x_right))
        join_decalaged_x_order = join_decalaged_x.argsort()
        x_decalaged = join_decalaged_x[join_decalaged_x_order]
        
        return x_decalaged, decalaged_x_left, decalaged_x_right, join_decalaged_x_order

    @staticmethod
    def __decalaged_momentums(x_decalaged,
                              decalaged_x_left,
                              decalaged_x_right,
                              join_decalaged_x_order,
                              momentum_diagram):
        #momentum_decalaged_diagram = np.concatenate((momentum_diagram, momentum_diagram, momentum_diagram))[join_decalaged_x_order]

        momentum_left = np.interp(x_decalaged, decalaged_x_left, momentum_diagram)
        momentum_right = np.interp(x_decalaged, decalaged_x_right, momentum_diagram)

        momentum_positive = np.max((momentum_left, momentum_right), axis=0)
        momentum_positive = np.where(momentum_positive<0, np.nan, momentum_positive)
        
        momentum_negative = np.min((momentum_left, momentum_right), axis=0)
        momentum_negative = np.where(momentum_negative>0, np.nan, momentum_negative)
        return momentum_positive, momentum_negative  
        
    @staticmethod
    def __join_momentum_peak(_momentum):
        momentum = _momentum.copy() 
        peaks, _ = find_peaks(np.absolute(momentum))
        for peak_index in 2*np.arange(len(peaks)//2):
            x_momentum_index = peaks[peak_index]
            next_x_momentum_index = peaks[peak_index+1]
            momentum[x_momentum_index:next_x_momentum_index] = momentum[x_momentum_index]
        return momentum

    def getMinimumAndMaximumSteelArea(self, x):
        """
            Returns tuple of minimum and maximum necessary steel area given the position.

                Call signatures:

                    concrete_beam.long_steel_bars_solution_info.getMinimumAndMaximumSteelArea(x)

                >>> concrete_beam.long_steel_bars_solution_info.getMinimumAndMaximumSteelArea(300)
                (2.76, 19.2)
                
            Parameters
            ----------
            x : number
                Define the position in cm.
            
        """
        _, beam_element = self.concrete_beam.getBeamElementInX(x)
        return LongSteelBar.getMinimumAndMaximumSteelArea(
            area = beam_element.section.area,
            fck = beam_element.material.fck
        )
        
        
    
    def getComercialSteelAreaDiagram(self, **options_diagram):
        """
            Returns comercial steel area diagram.
            Implements: minimum steel area, check maximum steel area and do not allow a single steel bar.
            Does not have the removal by step implemented here.

                Call signatures:

                    concrete_beam.long_steel_bars_solution_info.getComercialSteelAreaDiagram(division=1000)

                >>> x_decalaged, positive_areas_info, negative_areas_info = concrete_beam.long_steel_bars_solution_info.getComercialSteelAreaDiagram()
                >>> x_decalaged, positive_areas_info, negative_areas_info = concrete_beam.long_steel_bars_solution_info.getComercialSteelAreaDiagram(division=5000)
        """ 
        x_decalaged, momentum_positive, momentum_negative = timeit(self.verbose)(self.getDecalagedMomentumDesignDiagram)(**options_diagram)
        positive_areas_info = [self.getComercialSteelArea(x, m) for x, m in zip(x_decalaged, momentum_positive)]
        negative_areas_info = [self.getComercialSteelArea(x, m) for x, m in zip(x_decalaged, momentum_negative)]
        
        #positive_areas_info = timeit("positive_areas_info", self.verbose)([self.getComercialSteelArea(x, m) for x, m in zip(x_decalaged, momentum_positive)])
        #negative_areas_info = timeit("negative_areas_info", self.verbose)([self.getComercialSteelArea(x, m) for x, m in zip(x_decalaged, momentum_negative)])
        return x_decalaged, np.array(np.array(positive_areas_info).T), np.array(np.array(negative_areas_info).T)

    
    def getComercialSteelArea(self, x, momentum):
        """
            Returns comercial steel area given the position and momentum.
            Implements: minimum steel area, check maximum steel area and do not allow a single steel bar.
            Does not have the removal by step implemented here.
            Not recommended to use in loops.

                Call signatures:

                    concrete_beam.long_steel_bars_solution_info.getComercialSteelArea(x, momentum)

                >>> concrete_beam.long_steel_bars_solution_info.getComercialSteelArea(300, 2500)
                (6.0, 0.8, 3.0)
                
            Parameters
            ----------
            x : number
                Define the position in cm.
                
            momentum : number
                Define the momentum in kNcm.
            
        """ 
        min_area, max_area = self.getMinimumAndMaximumSteelArea(x)
        area = self.getSteelArea(x, momentum)
        
        # Implement minimun area in support
        
        if abs(area)>max_area: raise Exception("Too much steel needed in x={}, area needed is {}cmˆ2, but the maximum is {}cmˆ2".format(x, abs(area), max_area))
        
        if np.isnan(area): return np.repeat(np.nan, 3)
        if area>0 :
            area = max(min_area, area)
            possible_areas = self.available.table[:,2] > area
            if len(possible_areas)==0: raise Exception("There is not a possible available longitudinal steel bar. You should try to increase max_number of fc.AvailableLongConcreteSteelBar.")
            values = self.available.table[possible_areas][0]
        else:
            possible_areas = self.available.table[:,2] < area
            if len(possible_areas)==0: raise Exception("There is not a possible available longitudinal steel bar. You should try to increase max_number of fc.AvailableLongConcreteSteelBar.")
            values = self.available.table[possible_areas][-1]
        quantity, diameter, area = values
        return quantity, diameter, area
    
    def _anchorSteelBars(self, steel_bars, interspace_between_momentum):
        steel_bar_surface_type = self.concrete_beam.available_long_steel_bars.surface_type
        n1 = (2.25 if steel_bar_surface_type == "ribbed"
        else 1 if steel_bar_surface_type == "plain"
        else 1.4 if steel_bar_surface_type == "carved"
        else 0)
                
        for interspace in interspace_between_momentum:
            steel_bars_in_insterspace = LongSteelBars(steel_bars[(interspace == steel_bars.interspaces).sum(axis=1)== 2])
            major_steel_bar = steel_bars_in_insterspace[abs(steel_bars_in_insterspace.areas_accumulated) == abs(steel_bars_in_insterspace.areas_accumulated).max()][0]
            diameter = major_steel_bar.diameter
            begin, end = major_steel_bar.long_begin, major_steel_bar.long_end
            _, bar_element = self.concrete_beam.getBeamElementInX((begin+end)/2)
            
            is_in_the_beam_element = (self.x >= begin) &  (self.x <= end)
            positive_area_info = self.positive_areas_info[:, is_in_the_beam_element]
            negative_area_info = self.negative_areas_info[:, is_in_the_beam_element]

            positive_area_diagram = positive_area_info[2]
            negative_area_diagram = negative_area_info[2]
            positive_area_diagram = positive_area_diagram[~np.isnan(positive_area_diagram)]
            negative_area_diagram = negative_area_diagram[~np.isnan(negative_area_diagram)]
            
            As_calc = max(positive_area_diagram.min(initial=0), negative_area_diagram.min(initial=0),
                        positive_area_diagram.max(initial=0), negative_area_diagram.max(initial=0), key=abs)

            As_ef = major_steel_bar.area_accumulated

            # Can make a more precise calculus here.
            # 1 if (h < 60 and bar_transv_y <= 30) or
            # (h >= 60 and bar_transv_y <= h-30) else 0.7
            n2 = 1 if major_steel_bar.area > 0 else 0.7
            n3 = 1 if diameter < 3.2 else (13.2 - diameter)/10

            f_bd = n1 * n2 * n3 * bar_element.material.fctd

            # Check if hook is necessary
            alpha = 0.7 if (begin <= self.concrete_beam.x_begin+bar_element.material.c or end >= self.concrete_beam.x_end-bar_element.material.c) else 1

            lb = max(abs(diameter*major_steel_bar.fyd/(4*f_bd)), 25*diameter)
            lbmin = max(0.3*lb, 10*diameter, 10)
            lb_nec = max(alpha*lb*As_calc/As_ef, lbmin)

            if f_bd ==0: raise Exception("fbd as zero")

            steel_bars = steel_bars.changeProperty("long_begin",
                                                   function=lambda x:x-lb_nec,
                                                   conditional=lambda x:list(x.interspace)==list(interspace))
            steel_bars = steel_bars.changeProperty("long_end",
                                                   function=lambda x:x+lb_nec,
                                                   conditional=lambda x:list(x.interspace)==list(interspace))
            
        return steel_bars


    def _getBarsInInterspaces(self, x, areas_info, interspaceBetweenMomentum):
        quantities, diameters, _ = areas_info  
        bar_steel_removal_step = self.bar_steel_removal_step
        bar_steel_max_removal = self.bar_steel_max_removal
        
        bars = LongSteelBars()
        for interspace in interspaceBetweenMomentum:
            bars_interspace = LongSteelBars()
            times_removal_occurred = 0
            
            #commum interspace info
            interspace_start, interspace_end = interspace[0], interspace[1]
            is_in_interpace = (x > interspace_start) & (x<interspace_end)
            x_interspace = x[is_in_interpace]
            quantities_interspace = quantities[is_in_interpace]
            diameter = diameters[is_in_interpace][0]
            
            max_quantity_interspace = int(max(quantities_interspace))
            min_quantity_interspace = int(min(quantities_interspace))
                    
            for quantity in range(1, max_quantity_interspace+1):
                if np.isin(quantity, bars_interspace.quantities_accumulated): continue
                x_same_quantity = x_interspace[quantities_interspace == quantity]
                if len(x_same_quantity)>0:
                    # Just removing bars according to bar_steel_removal_step
                    reminder = (quantity-min_quantity_interspace)%bar_steel_removal_step
                    quantity_accumulated = min(quantity - reminder + (reminder>0)*self.bar_steel_removal_step, max_quantity_interspace)
                    new_reminder = (quantity_accumulated-min_quantity_interspace)%bar_steel_removal_step
                    new_quantity = bar_steel_removal_step-new_reminder if min_quantity_interspace!=quantity_accumulated else min_quantity_interspace
                    
                    removal_limit_reached = times_removal_occurred>=bar_steel_max_removal-1
                    
                    
                    quantity = new_quantity if not removal_limit_reached else new_quantity+max_quantity_interspace-quantity_accumulated
                    quantity_accumulated = quantity_accumulated if not removal_limit_reached else max_quantity_interspace
        
                    area_accumulated = self.available.diameters_to_area[abs(diameter*10)]*quantity_accumulated*(1 if diameter>0 else -1)
                    fyd = self.available.fyd
                    area = self.available.diameters_to_area[abs(diameter*10)]*quantity*(1 if diameter>0 else -1)
                    long_begin = min(x_same_quantity)
                    long_end = max(x_same_quantity)
                    length = long_end-long_begin
                    cost = quantity*length*self.available.cost_by_meter[abs(diameter*10)]/100

                    new_bar = LongSteelBar(
                                    long_begin=long_begin,
                                    long_end=long_end,
                                    quantity=quantity,
                                    diameter=diameter,
                                    quantity_accumulated = quantity_accumulated,
                                    interspace=(interspace_start, interspace_end),
                                    area = area,
                                    area_accumulated = area_accumulated,
                                    fyd = fyd,
                                    length=length,
                                    cost = cost
                                    )
                    
                    bars_interspace.add(new_bar)
                    times_removal_occurred+=1
                    
                    if removal_limit_reached:
                        break
                    
            bars.add(bars_interspace)

        return bars
    
    #con be static
    def _getInterspaceBetweenMomentum(self, x, area):
        """
            Return an array and each row represents a interspace.
            Element row[0] is the begin of interspace and row[1], the end.
        """
        previous_y=np.nan
        interspace = np.array([0, 0])
        if sum(np.isfinite(area)) == 0 : return []
        for x_u,y in zip(x, area):
            if np.isnan(previous_y) and not np.isnan(y):
                begin = x_u
            elif np.isnan(y) and not np.isnan(previous_y):
                end = x_u
                interspace = np.vstack([interspace, [begin, end]])
            previous_y = y
        interspace = np.vstack([interspace, [begin, x[-1]]])[1:]
        return interspace


    def getSteelArea(self, x, momentum):
            """
                #only working with rectangle section
                Returns necessary steel area given the position and momentum.

                    Call signatures:

                        concrete_beam.long_steel_bars_solution_info.getSteelArea(x, momentum)

                    >>> concrete_beam.long_steel_bars_solution_info.getSteelArea(10, 2500)
                    0.903512040037519
                    
                Parameters
                ----------
                x : number
                    Define the position in cm.
                    
                momentum : number
                    Define the momentum in kNcm.
                
            """ 
            #only working with rectangle section
            _, single_beam = self.concrete_beam.getBeamElementInX(x)
            return LongSteelBar.getSteelArea(section=single_beam.section,
                                            material=single_beam.material,
                                            steel=self.available,
                                            momentum=momentum)
        
        
    def getSteelAreaDiagram(self, **options_diagram):
        """
            Returns necessary steel area diagram.

                Call signatures:

                    concrete_beam.long_steel_bars_solution_info.getSteelAreaDiagram(division=1000)

                >>> x_decalaged, positive_areas, negative_areas = concrete_beam.long_steel_bars_solution_info.getSteelAreaDiagram()
                >>> x_decalaged, positive_areas, negative_areas = concrete_beam.long_steel_bars_solution_info.getSteelAreaDiagram(division=20)
            
        """ 
        x_decalaged, momentum_positive, momentum_negative = self.getDecalagedMomentumDesignDiagram(**options_diagram)
        positive_areas = [self.getSteelArea(x, m) for x, m in zip(x_decalaged, momentum_positive)]
        negative_areas = [self.getSteelArea(x, m) for x, m in zip(x_decalaged, momentum_negative)]
        return x_decalaged, np.array(positive_areas), np.array(negative_areas)
    
    
    def getDecalagedLength(self, beam_element):
        """
            Returns decalaged length of a beam element.

                Call signatures:

                    concrete_beam.long_steel_bars_solution_info.getDecalagedLength(beam_element)
        """ 
        bw = beam_element.section.bw
        d = beam_element.section.maximum_steel_height
        _, s = self.concrete_beam.getShearDiagram(x_begin=beam_element.n1.x, x_end=beam_element.n2.x)
        alpha = radians(self.concrete_beam.available_transv_steel_bars.inclination_angle)
        if alpha==radians(90):
            return 0.5*d
        vsd_max = max(abs(s))
        fctd = beam_element.material.fctd
        v_c0 = 0.6*fctd*bw*d
        al_formula = min(d*((vsd_max*(1+tan(alpha)**(-1)))/(2*(vsd_max-v_c0))-tan(alpha)**(-1)), d)
        if vsd_max <= v_c0:
            al = d
        elif alpha==radians(45):
            al = max(al_formula, 0.2*d)
        return al
    
    
    


import numpy as np
from fconcrete.helpers import to_unit

class Node:
    def __init__(self, x, condition_boundary, length=0):
        """
            Represents a generical node.
            Node is the delimitation for a beam_element.
        """
        x = to_unit(x, "cm")
        length = to_unit(length, "cm")
        self.x = x
        self.condition_boundary = condition_boundary
        self.length = length
        
    @classmethod
    def SimpleSupport(cls, x, length=0):
        """
            Represents a node with vertical displacement equal to zero.
            But it allows rotation.
            
            Call signatures:

                    fc.Node.SimpleSupport(x, length=0)

                >>> simple_support_1 = fc.Node.SimpleSupport(100)
                >>> simple_support_2 = fc.Node.SimpleSupport('1m')
                >>> repr(simple_support_1) == repr(simple_support_2)
                True

            Parameters
            ----------
            x : number or str
                Position of the node. If it is a number, default unit is cm, but also [length] unit can be given. Example:
                '20m', '10dm', etc
            
            length : number or str, optional
                Length of the node if applicable. If it is a number, default unit is cm, but also [length] unit can be given. Example:
                '20m', '10dm', etc.
                Default is 0.
        """
        return cls(x, [0, 1], length)
    
    @classmethod
    def Free(cls, x):
        """
            Represents a node with vertical displacement and rotation.
            
            Call signatures:

                    fc.Node.Free(x)

                >>> free_node_1 = fc.Node.Free(100)
                >>> free_node_2 = fc.Node.Free('1m')
                >>> repr(free_node_1) == repr(free_node_2)
                True
            
            Parameters
            ----------
            x : number or str
                Position of the node. If it is a number, default unit is cm, but also [length] unit can be given. Example:
                '20m', '10dm', etc
        """
        return cls(x, [1, 1])
    
    @classmethod
    def MiddleNode(cls, x):
        """
            Represents a node with vertical displacement and rotation.
            
            Call signatures:

                    fc.Node.Free(x)

                >>> middle_node_1 = fc.Node.MiddleNode(100)
                >>> middle_node_2 = fc.Node.MiddleNode('1m')
                >>> repr(middle_node_1) == repr(middle_node_2)
                True
            
            Parameters
            ----------
            x : number or str
                Position of the node. If it is a number, default unit is cm, but also [length] unit can be given. Example:
                '20m', '10dm', etc
        """
        return cls(x, [1, 1])
    
    @classmethod
    def Crimp(cls, x, length=0):
        """
            Represents a node with vertical displacement and rotation equal to zero.
            
            Call signatures:

                    fc.Node.Crimp(x)

                >>> crimp_node_1 = fc.Node.Crimp(100)
                >>> crimp_node_2 = fc.Node.Crimp('1m')
                >>> repr(crimp_node_1) == repr(crimp_node_2)
                True
            
            Parameters
            ----------
            x : number or str
                Position of the node. If it is a number, default unit is cm, but also [length] unit can be given. Example:
                '20m', '10dm', etc
            
            length : number or str, optional
                Length of the node if applicable. If it is a number, default unit is cm, but also [length] unit can be given. Example:
                '20m', '10dm', etc.
                Default is 0.
        """
        return cls(x, [0, 0], length)
    
    def __repr__(self):
        return str(self.__dict__)+'\n'


class Nodes:
    def __init__(self, nodes):
        self.nodes = np.array(nodes)
        self.x = np.array([ node.x for node in nodes ])
        self.condition_boundary = np.array([ node.condition_boundary for node in nodes ])
    
    def __repr__(self):
        return str(self.nodes)
    
    def __getitem__(self, key):
        return self.nodes[key]
    
    def __len__(self):
        return len(self.nodes)
    


from fconcrete.Structural.Section import Section

class ConcreteSection(Section):
    
    @staticmethod
    def setSteelHeight(
        section,
        positive_steel_height = 0,
        negative_steel_height = 0
    ):
        
        if (positive_steel_height == 0 and negative_steel_height==0):
            p, n = section.height*0.8, section.height*0.8
            minimum, maximum = min(p, n), max(p, n)
        elif(positive_steel_height == 0 or negative_steel_height == 0):
            p = positive_steel_height if positive_steel_height else 0
            n = negative_steel_height if negative_steel_height else 0
            maximum = minimum = p + n
        else:
            p, n = positive_steel_height, negative_steel_height
            minimum, maximum = min(p, n), max(p, n)
            
        section.positive_steel_height = p
        section.negative_steel_height = n
        section.maximum_steel_height = maximum
        section.minimum_steel_height = minimum
        
        return section


from fconcrete import config, duplicated, Material, Beam, Load, Node, ConcreteBeam, BeamElement, Rectangle, Concrete, Section
e = config.e
from pytest import approx
import numpy as np
import os

def approx01(x):
    return approx(x, abs=0.1)


def compare(beam, name):
    shear_info, momentum_info, displacement_info, rotation_info = get_ftool_fconcrete_comparisson(beam,
                                    r"tests/structural/{}/shear.txt".format(name),
                                    r"tests/structural/{}/momentum.txt".format(name),
                                    r"tests/structural/{}/displacement.txt".format(name))
    
    shear_diagram_v47, shear_fconcrete = shear_info
    momentum_diagram_v47, momentum_fconcrete = momentum_info
    displacement_diagram_v47, displacement_fconcrete = displacement_info
    rotation_diagram_v47, rotation_fconcrete = rotation_info
    
    assert shear_diagram_v47 == approx(shear_fconcrete, abs=0.000001)
    assert momentum_fconcrete == approx(momentum_diagram_v47, abs=0.0001)
    assert rotation_diagram_v47 == approx(rotation_fconcrete, abs=0.0000001)
    assert displacement_diagram_v47 == approx(displacement_fconcrete, abs=0.000001)

def get_ftool_fconcrete_comparisson(beam, file_shear, file_momentum, file_displacement):
    
    x_shear, shear_diagram_v47 = np.loadtxt(file_shear).T             
    not_duplicated_x = ~duplicated(x_shear)
    x_shear = x_shear[not_duplicated_x][1:-1]
    shear_diagram_v47 = shear_diagram_v47[not_duplicated_x][1:-1]
    shear_fconcrete = beam.getInternalShearStrength(x_shear)


    x_momentum, momentum_diagram_v47 = np.loadtxt(file_momentum).T
    not_duplicated_x = ~duplicated(x_momentum)
    x_momentum = x_momentum[not_duplicated_x][1:-1]
    momentum_diagram_v47 = momentum_diagram_v47[not_duplicated_x][1:-1]
    momentum_fconcrete = beam.getInternalMomentumStrength(x_momentum)

    x_ftool, _, displacement_diagram_v47, rotation_diagram_v47 = np.loadtxt(file_displacement).T
    not_duplicated_x = ~duplicated(x_ftool)

    not_duplicated_x_displacement = not_duplicated_x & ~(displacement_diagram_v47 == 0)
    x_displacement = x_ftool[not_duplicated_x_displacement]
    displacement_diagram_v47 = displacement_diagram_v47[not_duplicated_x_displacement]
    displacement_fconcrete = beam.getDisplacement(x_displacement)*10

    x_rotation = x_ftool[not_duplicated_x]
    rotation_diagram_v47 = rotation_diagram_v47[not_duplicated_x]
    rotation_fconcrete = beam.getRotation(x_rotation)

    return (shear_diagram_v47, shear_fconcrete), (momentum_diagram_v47, momentum_fconcrete), (displacement_diagram_v47, displacement_fconcrete), (rotation_diagram_v47, rotation_fconcrete) 
    

def test_v47():
    material = Material(E='27000 MPa', poisson=1, alpha=1)
    section = Rectangle(25,44.6)

    f1 = Load.UniformDistributedLoad(-0.1622, x_begin=0, x_end=113)
    f2 = Load.UniformDistributedLoad(-0.4994, x_begin=113, x_end=583)
    f3 = Load.UniformDistributedLoad(-0.4196, x_begin=583, x_end=1188)

    n1 = Node.SimpleSupport(x=0)
    n2 = Node.SimpleSupport(x=113)
    n3 = Node.SimpleSupport(x=583)
    n4 = Node.SimpleSupport(x=1188)

    bar1 = BeamElement([n1, n2], section, material)
    bar2 = BeamElement([n2, n3], section, material)
    bar3 = BeamElement([n3, n4], section, material)

    beam = Beam(
        loads = [f1, f2, f3],
        beam_elements = [bar1, bar2, bar3]
    )
    compare(beam=beam, name="v47")
    

def test_free_crimped():
    material = Material(E='27000 MPa', poisson=1, alpha=1)
    section = Rectangle(25,44.6)

    f1 = Load.PontualLoad(-5, x=29)
    f2 = Load.UniformDistributedLoad(-0.1622, x_begin=0, x_end=113)

    n1 = Node.Free(x=0)
    n2 = Node.Crimp(x=113)

    bar1 = BeamElement([n1, n2], section, material)

    beam = Beam(
        loads = [f1, f2],
        beam_elements = [bar1]
    )
    compare(beam=beam, name="free_crimped")

def test_crimped_free():
    material = Material(E='27000 MPa', poisson=1, alpha=1)
    section = Rectangle(25,44.6)

    f1 = Load.UniformDistributedLoad(-0.1622, x_begin=0, x_end=113)

    n1 = Node.Crimp(x=0)
    n2 = Node.Free(x=113)

    bar1 = BeamElement([n1, n2], section, material)

    beam = Beam(
        loads = [f1],
        beam_elements = [bar1]
    )
    compare(beam=beam, name="crimped_free")
    
    
def test_crimped_simple_supported():
    material = Material(E='27000 MPa', poisson=1, alpha=1)
    section = Rectangle(25,44.6)

    f1 = Load.UniformDistributedLoad(-0.1622, x_begin=0, x_end=113)

    n1 = Node.Crimp(x=0)
    n2 = Node.SimpleSupport(x=113)

    bar1 = BeamElement([n1, n2], section, material)

    beam = Beam(
        loads = [f1],
        beam_elements = [bar1]
    )
    compare(beam=beam, name="crimped_simple_supported")
    
    
    #for x, ft, fc in zip(x_rotation, rotation_diagram_v47, rotation_fconcrete):
    #if ft != approx(fc, abs=0.000001):
    #    print(x, ft, fc)


"""
    Code for structural calculus using concrete.
"""
from .AvailableMaterials import *
from .Concrete import Concrete
from .TransvSteelBar import *
from .LongSteelBar import *
from .Analysis import *
from .ConcreteBeam import ConcreteBeam
from .ConcreteSection import *


"""
    Code for structural calculus. Not related to any specific material.
    Uses FEM (Finite Element Method) to define the efforts.
"""

from fconcrete.Structural.Beam import *
from fconcrete.Structural.Load import *
from fconcrete.Structural.Node import *
from fconcrete.Structural.Section import *
from fconcrete.Structural.Material import *
from fconcrete.Structural.BeamElement import *


from fconcrete.helpers import cond, integrate, to_unit
        
        
class Material():
    """
        Define the class for the material.
        
        Attributes
        ----------
        E : number
            Represent the Young Modulus (E) in kN/cmˆ2.
        
        poisson : number
            Poisson's ratio is a measure of the Poisson effect, that describes the expansion of a material in directions perpendicular to the direction of compression.
             
        alpha : number   
            Coefficient of thermal expansion which is the relative expansion (also called strain) divided by the change in temperature.
    """
    def __init__(self, E, poisson, alpha):
        """
            Define a material and its properties.

                Call signatures:

                    fc.Material(E, poisson, alpha)

                >>> generical_material = fc.Material(20000, 0.3, 10**(-6))
                >>> generical_material.E
                2000

            Parameters
            ----------
            E : number or str
                Represent the Young Modulus (E). If it is a number, default unit is kN/cmˆ2, but also [force]/[distance]**2 unit can be given. Example:
                '20000MPa', '10N/mmˆ2', etc
                
            poisson : number
                Poisson's ratio is a measure of the Poisson effect, that describes the expansion of a material in directions perpendicular to the direction of compression.
             
            alpha : number   
                Coefficient of thermal expansion which is the relative expansion (also called strain) divided by the change in temperature.
        """
        self.E = to_unit(E, "kN/cm**2")
        self.poisson = poisson
        self.alpha = alpha
    
unitary_material = Material(10**6, 1, 1)


import numpy as np
from fconcrete.Structural.Node import Node, Nodes
from fconcrete.Structural.Section import unitary_section
from fconcrete.Structural.Material import unitary_material
import copy

class BeamElement:
    """
        Class that defines a primitive elements of a beam.
    """
    def __init__(self, nodes, section=unitary_section, material=unitary_material):
        """
            Define the beam_elements that, together, makes the whole Beam. 
            
                Call signatures:

                    fc.BeamElement(nodes, section=unitary_section, material=unitary_material)
            
            Parameters
            ----------
            nodes : list of Node
                list of fc.Node to represent the delimitation for the beam_element.

            section : Section, optional
                Define the section that are going to make the beam_element.
                Default fc.unitary_section.
                
            material : Material, optional
                Define a material and its properties.
                Default fc.unitary_material.
        """
        self.section = section
        self.material = material
        self.x = nodes
        self.E = material.E
        self.I = section.I
        self.n1 = nodes[0]
        self.n2 = nodes[1]
        self.length = nodes[1].x - nodes[0].x
        self.flexural_rigidity = material.E*section.I
        
    def get_matrix_rigidity_unitary(self):
        """
            Returns the unitary rigidity matrix. 
        """
        return self.flexural_rigidity/(self.length**3)*np.array([
                        [12, 6*self.length, -12, 6*self.length],
                        [6*self.length, 4*self.length**2, -6*self.length, 2*self.length**2],
                        [-12, -6*self.length, 12, -6*self.length],
                        [6*self.length, 2*self.length**2, -6*self.length, 4*self.length**2]
                    ])
    
    @classmethod
    def get_efforts_from_bar_element(cls, beam_element, load):
        """
            Get the efforts coused by the load in a double crimped beam element.
            
            Parameters
            ----------
            distance_a : number
                Distance, in cm, from the left node to the force.
        """
        force = load.force
        length = beam_element.length
        order = load.order    
        distance_a = load.x - beam_element.x[0].x
                
        if distance_a>length: raise Exception("Distance from node cannot exceed the beam size.")
        distance_b = length - distance_a

        if order == 0:
            ma= force*distance_a*distance_b**2/length**2
            mb= -force*distance_b*distance_a**2/length**2 
            ra = (force*distance_b+ma+mb)/length
        elif order == 1:
            ma= load.q*length**2/12
            mb= -ma
            ra = load.q*length/2
        
        return -np.array([ra, ma, force-ra, mb])
 
    def split(self, x):
        """
            Split a beam_element in two. The node in x is considered a Middle Node.
            
            Parameters
            ----------
            x : number
                Distance, in cm, from the left node to the split point.
        """
        if x >= self.n2.x or x <= self.n1.x: return [self]
        n_intermediate = Node.MiddleNode(x=x)
        bar1 = BeamElement(nodes=[self.n1, n_intermediate], section=self.section, material=self.material)
        bar2 = BeamElement(nodes=[n_intermediate, self.n2], section=self.section, material=self.material)
        return [bar1, bar2]
        
    def __repr__(self):
        return str(self.__dict__)


class BeamElements:
    """
        Class that defines a primitive elements of a beam list with easy to work properties and methods.
    """
    def __init__(self, bar_elements):
        self.bar_elements = np.array(bar_elements)
        self.materials = np.array([ bar_element.material for bar_element in bar_elements ])
        self.sections = np.array([ bar_element.section for bar_element in bar_elements ])
        self.x_start = np.array([ bar_element.n1.x for bar_element in bar_elements ])
        self.x_end = np.array([ bar_element.n2.x for bar_element in bar_elements ])
        self.length = np.array([ bar_element.length for bar_element in bar_elements ])
        self.flexural_rigidity = np.array([ bar_element.flexural_rigidity for bar_element in bar_elements ])
        self.nodes = Nodes(np.concatenate((
                [ beam.n1 for beam in bar_elements ],
                [bar_elements[-1].n2 ] 
        )))
        condition_boundary = np.vstack((
            [ beam.n1.condition_boundary for beam in bar_elements ],
            bar_elements[-1].n2.condition_boundary
        ))
        self.condition_boundary = (condition_boundary.reshape((1, condition_boundary.size))==1)[0]
    
    @classmethod
    def create(cls, beam_elements):
        """
            Recommended way to create a BeamElements class.
        """
        if str(type(beam_elements)) == "<class 'fconcrete.Structural.BeamElement.BeamElements'>": return beam_elements
        beam_elements = np.array(beam_elements)
        x_start = np.array([ beam_element.n1.x for beam_element in beam_elements ])
        bar_sort_position = np.argsort(x_start)
        return cls(beam_elements[bar_sort_position])
    
    def split(self, x):
        """
            Similar to BeamElement.split, but can guess what element of the array is going to be splited.
        """
        new_beams = np.array([])
        for bar in self.bar_elements:
            new_beams = np.concatenate((new_beams, bar.split(x)))
        return BeamElements(new_beams)

    def changeProperty(self, prop, function, conditional=lambda x:True):
        """
            Change all properties of the beam elements in a single function.
        """
        beam_elements = copy.deepcopy(self)
        for previous_beam_element in beam_elements:
            if conditional(previous_beam_element):
                current_attribute_value = getattr(previous_beam_element, prop)
                setattr(previous_beam_element, prop, function(current_attribute_value)) 
        return BeamElements(beam_elements.bar_elements)

    def __repr__(self):
        return str(self.__dict__)    
    
    def __getitem__(self, key):
        return self.bar_elements[key]
    
    def __len__(self):
        return len(self.bar_elements)
    


from pytest import approx
import numpy as np
import os
import fconcrete as fc
import matplotlib.pyplot as plt
np.set_printoptions(precision=3, suppress=True, linewidth=3000)

def approx01(x):
    return approx(x, abs=0.1)

def create_concrete_beam():
    material = fc.Concrete(fck='30 MPa', aggressiveness=2)
    section = fc.Rectangle(25,60)

    #Design
    f1 = fc.Load.UniformDistributedLoad(-0.1622, x_begin=0, x_end=113)
    f2 = fc.Load.UniformDistributedLoad(-0.4994, x_begin=113, x_end=583)
    f3 = fc.Load.UniformDistributedLoad(-0.4196, x_begin=583, x_end=1188)

    n1 = fc.Node.SimpleSupport(x=0, length=20)
    n2 = fc.Node.SimpleSupport(x=113, length=20)
    n3 = fc.Node.SimpleSupport(x=583, length=20)
    n4 = fc.Node.SimpleSupport(x=1188, length=20)

    bar1 = fc.BeamElement([n1, n2], section, material)
    bar2 = fc.BeamElement([n2, n3], section, material)
    bar3 = fc.BeamElement([n3, n4], section, material)

        
    beam = fc.ConcreteBeam(
        loads = [f1, f2, f3],
        beam_elements = [bar1, bar2, bar3],
        bar_steel_max_removal = 2,
    )
    return beam
    
def test_create_concrete_beam():
    beam = create_concrete_beam()
    assert beam.processing_time>0
    


import numpy as np
from fconcrete import config as c
import matplotlib.pyplot as plt
import time

_Q = c._Q

def cond(x, singular=False, order=0):
    """
    If It is singular, return 1 if x>0 else 0.
    If It is not singular, return x**order if x>0 else 0
    """
    if singular:
        return 1 if x>0 else 0
    return x**order if x>0 else 0

def integrate(f, a, b, N=100):
    """
    Integrate f from a to b in N steps
    """
    x = np.linspace(a, b, N)
    y = np.apply_along_axis(f, 0, np.array([x]))
    return np.trapz(y, dx=(b-a)/(N-1))

def duplicated(array):
    """
    Check if it is duplicated.
    """
    s = np.sort(array, axis=None)
    duplicated = s[:-1][s[1:] == s[:-1]]
    return np.isin(s, duplicated)

def to_unit(input, expected_unit, return_unit=False):
    """
        Convert between unities according to expected_unit and return_unit.

            Call signatures:

                fc.helpers.to_unit(input, expected_unit, return_unit=False)

            >>> unit1 = fc.helpers.to_unit("10cm", "m")
            >>> unit1
            0.1
            
            >>> unit2 = fc.helpers.to_unit(20, "m", return_unit="cm")
            >>> unit2
            2000.0
            
        Parameters
        ----------
        input : number or str
            Represents the input unit of the user.
        
        expected_unit : str
            The expected unit to be given. Useful when input is a number.
            
        return_unit : `bool`, optional
            The desired unit to return

    """
    try:
        input = float(input)
        value = _Q(input, expected_unit)
    except:
        pass
        try:
            value = _Q(input).to(expected_unit)
        except: raise Exception("String does not have valid format. See documentation.")
            
    if return_unit:
        return value.to(return_unit).magnitude
    return value.magnitude
        
def getAxis(xy0=(0,0), xy1=(0,0)):
    """
    Create axis with equal aspect. xy0 and xy1 represent the visible area.
    """
    x0, y0 = xy0
    x1, y1 = xy1
    fig, ax = plt.subplots()
    ax.set_aspect("equal")
    ax.plot([x0, x1], [y0, y1], color="None")
    return fig, ax

def timeit(do=True, name=""):
    """
    Decorator to print the time that the function has taken to execute.
    """
    def inner0(function):
        if not do: return function
        def inner(*args, **kw):
            start = time.time()
            val = function(*args, **kw)
            end = time.time()
            print("{} executed in {}s".format(function.__name__ if name == "" else name, end-start))
            return val
        return inner
    return inner0

# https://gist.github.com/snakers4/91fa21b9dda9d055a02ecd23f24fbc3d
def printProgressBar (iteration, total, prefix = '', suffix = '', decimals = 1, length = 100, fill = '█', printEnd = "\r"):
    """
    Call in a loop to create terminal progress bar
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end = printEnd)
    # Print New Line on Complete
    if iteration == total: 
        print()


from fconcrete.Structural.BeamElement import BeamElement, BeamElements
from fconcrete.Structural.Load import Load, Loads
from fconcrete.Structural.Node import Nodes
from fconcrete.helpers import cond
from fconcrete.config import e
import copy
import numpy as np
import warnings
import matplotlib.pyplot as plt

class Beam:
    """
        Structural Beam.\n
        Note for this documentation: "Not only the ones provided by the initial beam_Elements" means that internally the Beam automatically creates some nodes even if it was not created for the user initially. It happens in Load.x_begin and Load.x_end. 
        
        Attributes
        ----------
        U: list of number
            Displacement in the nodes.

        beam_elements: BeamElements
            BeamElements instance of beam_elements, not only the ones provided by the initial beam_Elements.

        beams_quantity: number
            Number of beam_elements, not only the ones provided by the initial beam_Elements.

        external_loads: Loads
            loads argument but used as a Loads class. Same as fconcrete.Structural.Load.Loads.create(loads).

        initial_beam_elements: array of BeamElement
            beam_elements argument used when the instance is created.

        length: number
            Length of the beam. Can also use len(beam).
        
        loads: Loads
            Loads instance with all efforts in the beam. Including the load given by the supports.
            
        nodal_efforts: list of number
            The nodal efforts that happens in all nodes, not only the ones provided by the initial beam_Elements.

        nodes: Nodes
            Nodes instance of the beam, not only the ones provided by the initial beam_Elements.
            
        x_begin: number
            Where the beam starts, in cm.

        x_end: number
            Where the beam ends, in cm.
    """
    def __init__(self, loads, beam_elements, **options):
        """
            Returns a concrete_beam element.
            
                Call signatures:

                    fc.Beam(loads, beam_elements, **options)

                >>> n1 = fc.Node.SimpleSupport(x=0)
                >>> n2 = fc.Node.SimpleSupport(x=400)
                >>> n3 = fc.Node.SimpleSupport(x=800)
                >>> beam_element_1 = fc.BeamElement([n1, n2])
                >>> beam_element_2 = fc.BeamElement([n2, n3])
                >>> load_1 = fc.Load.UniformDistributedLoad(-0.000001, x_begin=0, x_end=1)
                 
                >>> beam = fc.Beam(
                >>>     loads = [f1],
                >>>     beam_elements = [beam_element_1, beam_element_2],
                >>> )
            
            Parameters
            ----------
            loads : [Load]
                Define the loads supported for the beam.
            
            beam_elements : [BeamElement], optional
                Define the beam_elements that, together, makes the whole Beam. 
        """
        beam_elements = BeamElements.create(beam_elements)
        external_loads = Loads.create(loads)
        self.initial_beam_elements = beam_elements
        beam_elements = self._createIntermediateBeams(external_loads, beam_elements)
        self.x_begin, self.x_end = beam_elements.nodes[0].x, beam_elements.nodes[-1].x
        
        self.external_loads = external_loads
        self.beam_elements = beam_elements
        
        self.length = sum(beam_elements.length)
        self.beams_quantity = len(beam_elements)
        
        if options.get("solve_structural") != False:
            self.solve_structural()
            if options.get("solve_displacement") != False:
                self.solve_displacement()
        
    def solve_structural(self):
        """
            Starts the process of solution for the structural beam.
        """
        nodal_efforts = self._getSupportReactions()
        self.nodal_efforts = nodal_efforts
        nodes = Nodes(self.beam_elements.nodes)
        self.nodes = nodes
        loads = self.external_loads
        for index, node in enumerate(nodes.nodes):
            loads = loads.add(
                [Load(nodal_efforts[index*2], nodal_efforts[index*2+1], node.x, node.x)]
                )
        self.loads = loads
                   
    @staticmethod
    def _createIntermediateBeams(loads, bars):
        for load in loads.loads:
            bars = bars.split(load.x_begin)
            bars = bars.split(load.x_end)
        
        return bars

    def matrix_rigidity_global(self):
        """
            Returns the global rigidity matrix. Also known by the letter "K". 
        """
        matrix_rigidity_row = 2*self.beams_quantity+2
        matrix_rigidity_global = np.zeros(
            (matrix_rigidity_row, matrix_rigidity_row))
        for beam_n, beam in enumerate(self.beam_elements):
            matrix_rigidity_global[beam_n*2:beam_n*2+4,
                                   beam_n*2:beam_n*2+4] += beam.get_matrix_rigidity_unitary()
        return matrix_rigidity_global

    def _get_beams_efforts(self):
        beams_efforts = np.zeros(self.beams_quantity*4)

        for load in self.external_loads.loads:
            if (load.x == self.x_end):
                load.x = self.x_end - e
            force_beam, beam_element = self.getBeamElementInX(load.x)

            beams_efforts[4*force_beam:4*force_beam+4] += BeamElement.get_efforts_from_bar_element(
                beam_element,
                load
            )

        # juntar vigas separadas em um vetor por nó
        bn = self.beams_quantity
        b = 2*np.arange(0, 2*bn+2, 1)
        b[1::2] = b[0::2]+1
        b[1] = b[-1] = b[-2] = 0
        mult_b = b != 0

        a = 2*(np.arange(0, 2*bn+2, 1)-1)
        a[0] = 0
        a[1::2] = a[0::2]+1

        return beams_efforts[a] + beams_efforts[b]*mult_b

    def _getSupportReactions(self):
        condition_boundary = self.beam_elements.condition_boundary
        beams_efforts = self._get_beams_efforts()
        matrix_rigidity_global = self.matrix_rigidity_global()

        matrix_rigidity_global_determinable = matrix_rigidity_global[condition_boundary, :][:, condition_boundary]
        beams_efforts_determinable = beams_efforts[condition_boundary]

        U = np.zeros(len(condition_boundary))
        U[condition_boundary] = np.linalg.solve(matrix_rigidity_global_determinable, beams_efforts_determinable)
        self.U = U
        F = matrix_rigidity_global @ U

        return beams_efforts - F
    

    def getBeamElementInX(self, x):
        """
            Get the beam element in x (in cm).

                Call signatures:

                    beam.getBeamElementInX(x)
            
            Parameters
            ----------
            x : number
                Position in the beam, in cm.
            
            Returns
            -------
            index : int
                The order of the beam_element in the structure.
                
            beam_element
                beam_element located in x.
                
        """
        index = 0 if x<=self.x_begin else -1 if x>=self.x_end else np.where(
        np.array([node.x for node in self.beam_elements.nodes]) <= x)[0][-1]
        bar_element = self.beam_elements[index]
        return index, bar_element
        
    def getInternalShearStrength(self, x):
        """
            Get the internal shear strength in a position x (in cm) or multiple positions.

                Call signatures:

                    beam.getInternalShearStrength(x)
            
            Parameters
            ----------
            x : number or list of number
                Position in the beam, in cm.
            
            Returns
            -------
            shear : number or list of number
                The internal value of the shear strength in kN.
                
        """
        if isinstance(x, int) or isinstance(x, float):
            if x < self.x_begin or x > self.x_end:
                return 0
            f_value = 0
            for load in self.loads.loads:
                f_value += load.force * \
                    cond(x-load.x_begin, singular=True) if load.x_begin == load.x_end else 0
                f_value += load.q*cond(x-load.x_begin, order=load.order) - load.q*cond(
                    x-load.x_end, order=load.order) if load.x_begin != load.x_end else 0
            return f_value
        elif isinstance(x, np.ndarray) or isinstance(x, list):
            return np.array([ self.getInternalShearStrength(x_element) for x_element in x ])
        
    def getShearDiagram(self, **options):
        """
            Apply beam.getInternalShearStrength for options["division"] parts of the beam.
            
            Parameters
            ----------
            **options
                
                ``division``:
                    Number of divisions equally spaced (`int`).
                ``x_begin``:
                    Begin of the x_axis (`number`).
                ``x_end``:
                    End of the x_axis (`number`).
                
                
            Returns
            -------
            x : list of number
                The x position of the division in cm
            
            y : list of number
                The value of shear for each x.
        """
        return self._createDiagram(self.getInternalShearStrength, **options)

    def getInternalMomentumStrength(self, x):
        """
            Get the internal momentum strength in a position x (in cm) or multiple positions.

                Call signatures:

                    beam.getInternalMomentumStrength(x)
            
            Parameters
            ----------
            x : number or list of number
                Position in the beam, in cm.
            
            Returns
            -------
            x : list of number
                The x position of the division in cm
                
            momentum : number or list of number
                The internal value of the momentum strength in kNcm.
                
        """
        if isinstance(x, int) or isinstance(x, float):
            if x < self.x_begin or x > self.x_end:
                return 0
            f_value = 0
            for load in self.loads.loads:
                f_value += -load.momentum * \
                    cond(x-load.x_begin, singular=True) if load.x_begin == load.x_end else 0
                f_value += load.force * \
                    cond(x-load.x_begin, order=1) if load.order == 0 else 0
                f_value += (load.q*cond(x-load.x_begin, order=load.order+1) -
                            load.q*cond(x-load.x_end, order=load.order+1))/(load.order+1)
            return f_value
        elif isinstance(x, np.ndarray) or isinstance(x, list):
            return np.array([ self.getInternalMomentumStrength(x_element) for x_element in x ])
        
    def getMomentumDiagram(self, **options):
        """
            Apply beam.getInternalMomentumStrength for options["division"] parts of the beam.
            
            Parameters
            ----------
            **options
                
                ``division``:
                    Number of divisions equally spaced (`int`).
                ``x_begin``:
                    Begin of the x_axis (`number`).
                ``x_end``:
                    End of the x_axis (`number`).
            
            Returns
            -------
            x : list of number
                The x position of the division in cm
            
            y : list of number
                The value of momentum for each x.
        """
        return self._createDiagram(self.getInternalMomentumStrength, **options)
    
    def solve_displacement(self):
        """
            Starts the process of solution for the structural beam displacement.
        """
        nodes = self.initial_beam_elements.nodes
        null_displacement = nodes.x[nodes.condition_boundary[:, 0]==0]
        null_rotation = nodes.x[nodes.condition_boundary[:, 1]==0]
        
        x1 = null_displacement[0]
        rest1 = self.getDisplacement(x1)
        
        if len(null_displacement)>=2:
            x2 = null_displacement[-1]
            rest2 = self.getDisplacement(x2)
            c1 = -(rest1 - rest2)/(x1-x2)
            c2 = -rest1 - c1*x1
        elif len(null_rotation)>=1 and len(null_displacement)>=1:
            x2 = null_rotation[0]
            c1 = -self.getRotation(x2)

        c2 = -rest1 - c1*x1
        
        self._c1 = c1
        self._c2 = c2
    
    def getDisplacement(self, x):
        """
            Get the vertical displacement in a position x (in cm) or multiple positions.

                Call signatures:

                    beam.getDisplacement(x)
            
            Parameters
            ----------
            x : number or list of number
                Position in the beam, in cm.
            
            Returns
            -------
            displacement : number or list of number
                The vertical displacement of the beam in cm.
                
        """
        if isinstance(x, int) or isinstance(x, float):
            if x < self.x_begin or x > self.x_end: return 0
            f_value = 0
            _, single_beam_element_init = self.getBeamElementInX(x)
            
            for load in self.loads:
                _, single_beam_element = self.getBeamElementInX(x) #load.x_begin+e)
                l_value = -load.momentum * \
                    cond(x-load.x_begin, order=2)/2 if load.x_begin == load.x_end else 0
                l_value += load.force * \
                    cond(x-load.x_begin, order=3)/6 if load.order == 0 else 0
                l_value += (load.q*cond(x-load.x_begin, order=load.order+3) -
                            load.q*cond(x-load.x_end, order=load.order+3))/((load.order+1)*(load.order+2)*(load.order+3))
                f_value += l_value/single_beam_element.flexural_rigidity
            
            if not hasattr(self, "_c1"):
                return f_value*single_beam_element_init.flexural_rigidity
            
            return f_value + (self._c1*x+self._c2)/single_beam_element_init.flexural_rigidity
        
        elif isinstance(x, np.ndarray) or isinstance(x, list):
            return np.array([ self.getDisplacement(x_element) for x_element in x ])
        
        
    def getDisplacementDiagram(self, **options):
        """
            Apply beam.getDisplacement for options["division"] parts of the beam.
            
            Parameters
            ----------
            **options
                
                ``division``:
                    Number of divisions equally spaced (`int`).
                ``x_begin``:
                    Begin of the x_axis (`number`).
                ``x_end``:
                    End of the x_axis (`number`).
                
            Returns
            -------
            x : list of number
                The x position of the division in cm
            
            y : list of number
                The value of displacement for each x.
        """
        return self._createDiagram(self.getDisplacement, **options)
    
    def getRotation(self, x):
        """
            Get the rotation in a position x (in cm) or multiple positions.

                Call signatures:

                    beam.getRotation(x)
            
            Parameters
            ----------
            x : number or list of number
                Position in the beam, in cm.
            
            Returns
            -------
            rotation : number or list of number
                The rotation value of the momentum strength in rad.
                
        """
        if isinstance(x, int) or isinstance(x, float):
            if x < self.x_begin or x > self.x_end: return 0
            f_value = 0
            _, single_beam_element_init = self.getBeamElementInX(x)
            for load in self.loads:
                _, single_beam_element = self.getBeamElementInX(load.x_begin+e)
                l_value = -load.momentum * \
                    cond(x-load.x_begin, order=1) if load.x_begin == load.x_end else 0
                l_value += load.force * \
                    cond(x-load.x_begin, order=2)/2 if load.order == 0 else 0
                l_value += (load.q*cond(x-load.x_begin, order=load.order+2) -
                            load.q*cond(x-load.x_end, order=load.order+2))/((load.order+1)*(load.order+2))
                f_value += l_value/single_beam_element.flexural_rigidity
                
            if not hasattr(self, "_c1"):
                return f_value*single_beam_element.flexural_rigidity
            return f_value + self._c1/single_beam_element_init.flexural_rigidity
        
        elif isinstance(x, np.ndarray) or isinstance(x, list):
            return np.array([ self.getRotation(x_element) for x_element in x ])
    
    def getRotationDiagram(self, **options):
        """
            Apply beam.getRotation for options["division"] parts of the beam.
            
            Parameters
            ----------
            **options
                
                ``division``:
                    Number of divisions equally spaced (`int`).
                ``x_begin``:
                    Begin of the x_axis (`number`).
                ``x_end``:
                    End of the x_axis (`number`).
                
            Returns
            -------
            x : list of number
                The x position of the division in cm
            
            y : list of number
                The value of rotation for each x.
        """
        return self._createDiagram(self.getRotation, **options)
    
    def _createDiagram(self, function, division=1000, x_begin="begin", x_end="end"):
        x_begin = self.x_begin+e if x_begin=="begin" else x_begin
        x_end = self.x_end-e if x_end=="end" else x_end
        x = np.linspace(x_begin, x_end, division)
        y = np.array([function(x_i) for x_i in x])
        return x, y
    
    def plotMomentumDiagram(self, **options):
        """
            Simply applies the beam.getMomentumDiagram method results (x,y) to a plot with plt.plot(x, y).\n
            Also invert y axis.
            
            Parameters
            ----------
            **options
                
                ``division``:
                    Number of divisions equally spaced (`int`).
                ``x_begin``:
                    Begin of the x_axis (`number`).
                ``x_end``:
                    End of the x_axis (`number`).
                
        """
        x, y = self.getMomentumDiagram(**options)
        plt.gca().invert_yaxis()
        plt.plot(x, y)
        
    def plotShearDiagram(self, **options):
        """
            Simply applies the beam.getShearDiagram method results (x,y) to a plot with plt.plot(x, y).

            Parameters
            ----------
            **options
                
                ``division``:
                    Number of divisions equally spaced (`int`).
                ``x_begin``:
                    Begin of the x_axis (`number`).
                ``x_end``:
                    End of the x_axis (`number`).
                
        """
        x, y = self.getShearDiagram(**options)
        plt.plot(x, y)
        
    def plotDisplacementDiagram(self, **options):
        """
            Simply applies the beam.getDisplacementDiagram method results (x,y) to a plot with plt.plot(x, y).
        
            Parameters
            ----------
            **options
                
                ``division``:
                    Number of divisions equally spaced (`int`).
                ``x_begin``:
                    Begin of the x_axis (`number`).
                ``x_end``:
                    End of the x_axis (`number`).
                
        """
        x, y = self.getDisplacementDiagram(**options)
        plt.plot(x, y)
        
    def plotRotationDiagram(self, **options):
        """
            Simply applies the beam.getRotationDiagram method results (x,y) to a plot with plt.plot(x, y).
        
            Parameters
            ----------
            **options
                
                ``division``:
                    Number of divisions equally spaced (`int`).
                ``x_begin``:
                    Begin of the x_axis (`number`).
                ``x_end``:
                    End of the x_axis (`number`).
                
        """
        x, y = self.getRotationDiagram(**options)
        plt.plot(x, y)
    
    def __name__(self):
        return "Beam"
    
    def __repr__(self):
        return str(self.__dict__)

    def copy(self):
        """
            Makes a deep copy of the instance of Beam.
        """
        return copy.deepcopy(self)
    
    


from fconcrete.Structural import Material
from fconcrete.helpers import to_unit
from math import log

class Concrete(Material):
    """
        Define the Concrete to be used and all its properties.
        
        Attributes
        ----------
        fck : number
            Define the characteristic resistance of the concrete in kN/cmˆ2.
        
        E_ci : number
            Modulus of elasticity or initial tangent deformation module of concrete, always referring to the cordal module in kN/cmˆ2.
        
        E_cs : number
            Secant deformation module of concrete in kN/cmˆ2.
        
        fctm : number
            Average concrete tensile strength in kN/cmˆ2.
            
        fctk_inf : number
            Minimum value of direct tensile strength in kN/cmˆ2.
        
        fctk_sup : number
            Maximum value of direct tensile strength in kN/cmˆ2.
            
        fcd : number
            Minimum value of design direct tensile strength in kN/cmˆ2.

        c : number
            Concrete covering in cm.
        
        wk : number
            Characteristic crack opening in the concrete surface in cm.
    """
    def __init__(self, fck, aggressiveness, aggregate="granite"):
        """
            Define properties of the concrete.

                Call signatures:

                    fc.Concrete(fck, aggressiveness, aggregate='granite', **kwargs)

                >>> material = fc.Concrete(fck=30, aggressiveness=3)
                >>> round(material.fck)
                3

            Parameters
            ----------
            fck : number
                Define the characteristic resistance of the concrete.
                If it is a number, default unit is MPa, but also [force]/[length]**2 unit can be given. Example:
                '20kN/cm**2', '10Pa', etc
                
            aggressiveness : int
                Aggressiveness value from 1 (very low) to 4 (very height)

            aggregate : {'basalt', 'diabase', 'granite', 'gneiss', 'limestone', 'sandstone'}
                Aggregate type.
                
        """ 
        fck = to_unit(fck, "MPa")
        
        alpha_e = 0
        if aggregate in ['basalt', 'diabase']: alpha_e = 1.2
        if aggregate in ['granite', 'gneiss']: alpha_e = 1
        if aggregate in ["limestone"]: alpha_e = 0.9
        if aggregate in ["sandstone"]: alpha_e = 0.7
        if alpha_e == 0: raise Exception("Must select a valid aggregate: 'basalt', 'diabase', 'granite', 'gneiss', 'limestone' or 'sandstone'")
        
        if fck<20 or fck>90: raise Exception("Must select a valid fck value (between 20MPa and 90MPa)")
        if (fck>=20 and fck<=50): E_ci = alpha_e*5600*(fck**0.5)
        E_ci = alpha_e*21500*(fck/10+1.25)**(1/3) if fck>50 else E_ci
        alpha_i = min(0.8+0.2*fck/80, 1)
        E_cs = alpha_i*E_ci
        
        fctm = 0.3*(fck**(2/3)) if fck<=50 else 2.12*log(1+0.11*fck, 10)
        fctk_inf = 0.7*fctm
        fctk_sup = 1.3*fctm
        fcd = fck/1.4
        fctd = fctk_inf/1.4

        c = 2.5 if aggressiveness==1 else 3 if aggressiveness==2 else 4 if aggressiveness==3 else 5 if aggressiveness==4 else 0
        if c==0: raise Exception("Must select a valid aggressiveness value (between 1 and 4)")
        wk = 0.04 if aggressiveness==1 else 0.03 if aggressiveness in [2, 3] else 0.02 if aggressiveness==4 else 0
        
        self.fck = to_unit(fck, "MPa", "kN/cm**2")
        self.E_ci = to_unit(E_ci, "MPa", "kN/cm**2")
        self.E_cs = to_unit(E_cs, "MPa", "kN/cm**2")
        self.fctm = to_unit(fctm, "MPa", "kN/cm**2")
        self.fctk_inf = to_unit(fctk_inf, "MPa", "kN/cm**2")
        self.fctk_sup = to_unit(fctk_sup, "MPa", "kN/cm**2")
        self.fcd = to_unit(fcd, "MPa", "kN/cm**2")
        self.fctd = to_unit(fctd, "MPa", "kN/cm**2")
        self.c = c
        self.wk = wk
        
        Material.__init__(self, E_cs, 0.2, 10**(-5))


import numpy as np
from math import sin, tan,  pi
from fconcrete.StructuralConcrete.Concrete import Concrete

#https://loja.arcelormittal.com.br/vergalhao-ca50-soldavel-63mm/p
class AvailableLongConcreteSteelBar:
    """
        Define the available longitudinal steel bars. 
        You can set the available diameters, cost_by_meter, fyw, E, etc.
        See more information in fc.AvailableLongConcreteSteelBar() docstring.
        For example, AvailableLongConcreteSteelBar([8]) means:
        
        - 8mm diameter;
        - 0.5cmˆ2 area;
        - R$2.0575 by meter cost;
        - fyw equal to 50kN/cmˆ2;
        - Young Modulus (E) is 21000kN/cmˆ2;
        - Max number of steel in the section is 200;
        - Surface type is ribbed.
    """
    def __init__(self,
                 diameters=[6.3, 8, 10, 12.5, 16, 20, 25, 32],
                 diameters_to_area={
                        6.3: 0.315,
                        8: 0.5,
                        10: 0.8,
                        12.5: 1.25,
                        16: 2,
                        20: 3.15,
                        25: 5,
                        32: 8,
                    },
                 cost_by_meter={
                        6.3: 15.39/12,
                        8: 24.69/12,
                        10: 36.89/12,
                        12.5: 54.79/12,
                        16: 89.79/12,
                        20: 140.29/12,
                        25: 219.09/12,
                        32: 402.39/12,
                    },
                 fyw = 50,
                 E = 21000,
                 max_number=200,
                 surface_type="ribbed"):
        """
            Returns a AvailableLongConcreteSteelBar instance.
            
            Parameters
            ----------
            diameter : list of number, optional
                Possible list of diameter in cm.
            
            diameters_to_area : dict, optional
                For each diameter (key), set the value of the area.
                
            cost_by_meter : dict, optional
                For each diameter (key), set the of cost by meter.
                
            fyw : number, optional
                Define the characteristic resistance of the steel in kN/cmˆ2.
                
            E : number, optional
                Define the Young Modulus (E) in kN/cmˆ2
                Default is 21000kN/cmˆ2.
            
            max_number : int, optional
                Max number of steel in the section.
                Default is 200.
            
            surface_type : {'ribbed', 'plain', 'carved'}, optional
                Surface type of the steel.
                Default is ribbed.
        """ 
        try:
            areas = [diameters_to_area[diameter] for diameter in diameters]
        except:
            raise Exception("Must provide a valid diameter to area dict")
        fyd = fyw/1.15
        diameters = np.array(diameters)
        areas = np.array(areas)
        diameters_loop = np.tile(diameters/10, max_number-1)
        # Single steel bar use not allowed, that is why range starts at 2
        areas_loop = np.concatenate([ areas*(i) for i in range(2, max_number+1)])
        number_of_bars = areas_loop/np.tile(areas,max_number-1)
        table_of_positive_steel = np.stack((number_of_bars, diameters_loop, areas_loop), axis=1)
        
        table_of_negative_steel = np.stack((number_of_bars, -diameters_loop, -1*areas_loop), axis=1)
        table_of_positive_and_negative_steel = np.vstack([table_of_negative_steel, table_of_positive_steel])
        
        table = np.array(table_of_positive_and_negative_steel[
            table_of_positive_and_negative_steel[:,2].argsort()
        ])
        self.table = table
        self.fyw = fyw
        self.fyd = fyd
        self.E = E
        self.diameters = diameters
        self.diameters_to_area = diameters_to_area
        self.surface_type = surface_type
        self.cost_by_meter = cost_by_meter
        
        
class AvailableTransvConcreteSteelBar:
    """
        Define the available transversal steel bars. 
        You can set the available diameters, cost_by_meter, fyw, E, etc.
        See more information in fc.AvailableTransvConcreteSteelBar docstring.
        Default AvailableTransvConcreteSteelBar([8]) which means:
        
        - 8mm diameter;
        - 0.5cmˆ2 area;
        - R$2.0575 by meter cost;
        - The longitudinal space between transversal steel are multiple of 5;
        - fyw equal to 50kN/cmˆ2;
        - Transversal bar inclination angle of 90 degrees;
        - Tilt angle of compression struts of 45 degrees.
    """
    def __init__(self,
                 diameters=[6.3, 8, 10, 12.5, 16, 20, 25, 32],
                 diameters_to_area={
                        6.3: 0.315,
                        8: 0.5,
                        10: 0.8,
                        12.5: 1.25,
                        16: 2,
                        20: 3.15,
                        25: 5,
                        32: 8,
                    },
                 cost_by_meter={
                        6.3: 15.39/12,
                        8: 24.69/12,
                        10: 36.89/12,
                        12.5: 54.79/12,
                        16: 89.79/12,
                        20: 140.29/12,
                        25: 219.09/12,
                        32: 402.39/12,
                    },
                 space_is_multiple_of=[5],
                 fyw = 50,
                 inclination_angle = 90,
                 ):
        """
            Returns a AvailableLongConcreteSteelBar instance.
            
            Parameters
            ----------
            diameter : list of number, optional
                Possible list of diameter in cm.
            
            diameters_to_area : dict, optional
                For each diameter (key), set the value of the area.
                
            cost_by_meter : dict, optional
                For each diameter (key), set the of cost by meter.
                
            space_is_multiple_of : list of number, optional
                The longitudinal spaces between transversal steel is multiple of the number of space_is_multiple_of list.
                Default is [5].
            
            fyw : number, optional
                Define the characteristic resistance of the steel in kN/cmˆ2.
                
            inclination_angle : number, optional
                Transversal bar inclination angle in degrees.
                Default is 90 degrees.
        """ 
            
        try:
            areas = [diameters_to_area[diameter] for diameter in diameters]
        except:
            raise Exception("Must provide a valid diameter to area dict")
        
        possible_spaces = np.array([])
        for multiple in space_is_multiple_of:
            possible_spaces = np.concatenate((possible_spaces,multiple*np.arange(1,30)))
        possible_spaces = np.unique(possible_spaces[possible_spaces <= 30])
        
        areas = [ diameters_to_area[diameter] for diameter in diameters ]
        diameters_loop = np.tile(diameters, len(possible_spaces))/10
        spaces_loop = np.repeat(possible_spaces, len(diameters))
        areas_loop = 2*np.tile(areas, len(possible_spaces))

        table = np.transpose([
            diameters_loop,
            spaces_loop,
            areas_loop,
            areas_loop/spaces_loop
        ])

        fyd = fyw/1.15
        diameters = np.array(diameters)
        
        self.fyw = fyw
        self.fyd = fyd
        self.table = table[table[:,3].argsort()]
        
        self.diameters = diameters
        self.diameters_to_area = diameters_to_area
        self.cost_by_meter = cost_by_meter
        self.inclination_angle = inclination_angle
        
#https://servicos.compesa.com.br/wp-content/uploads/2016/02/TABELA_COMPESA_2016_SEM_DESONERACAO_E_SEM_ENCARGOS_COMPLEMENTARES.pdf
class AvailableConcrete():
    """
        Define the available concrete. 
        You can set the available fck, cost_by_m3, aggressiveness and aggregate.
        See more information in fc.AvailableConcrete docstring.
        For example, AvailableConcrete() means:
        
        - 30 MPa;
        - R$353.30 by meterˆ3;
        - The aggressiveness is 3;
        - Aggregate is granite;
        - Biggest aggregate dimension is 1.5cm.
    """
    def __init__(self,
                 fck=30,
                 cost_by_m3=None,
                 aggressiveness=3,
                 aggregate='granite',
                 biggest_aggregate_dimension=1.5
                 ):
        """
            Returns a AvailableLongConcreteSteelBar instance.
            
            Parameters
            ----------
            fck : number
                Define the characteristic resistance of the concrete.
                If it is a number, default unit is MPa, but also [force]/[length]**2 unit can be given. Example:
                '20kN/cm**2', '10Pa', etc
            
            cost_by_m3 : number, optional
                Cost by mˆ3 of the concrete.
                If fck is 25, 30, 35 or 40, this value by default is set to 331.65, 353.30, 373.21, 385.36, respectively.
                
            aggressiveness : int
                Aggressiveness value from 1 (very low) to 4 (very height)

            aggregate : {'basalt', 'diabase', 'granite', 'gneiss', 'limestone', 'sandstone'}
                Aggregate type.
                
            biggest_aggregate_dimension : number, optional
                Maximum dimension characteristic of the biggest aggregate, in cm.
                Default value is 1.5.
        """ 
        if cost_by_m3 == None:
            try:
                cost_by_m3_dict = {
                    25: 331.65,
                    30: 353.30,
                    35: 373.21,
                    40: 385.36
                }
                cost_by_m3 = cost_by_m3_dict[fck]
            except:
                raise Exception("Please, provide cost for concrete with {}MPa".format(fck))
            
        self.fck = fck
        self.cost_by_m3 = cost_by_m3
        self.material = Concrete(str(fck) + " MPa", aggressiveness, aggregate)
        self.biggest_aggregate_dimension = biggest_aggregate_dimension
        
        
        
def solve_cost(concrete_beam, decimal_numbers = 2):
    cost_table = [["Material", "Price", "Quantity", "Unit", "Commentary", "Is Subtotal"]]
    # Concrete
    total_concrete_cost = 0
    for beam_element in concrete_beam.initial_beam_elements:
        volume = beam_element.section.area*beam_element.length/1000000
        concrete_cost = volume*concrete_beam.available_concrete.cost_by_m3
        total_concrete_cost += concrete_cost
        row = ["Concrete",
                round(concrete_cost, decimal_numbers),
                round(volume, decimal_numbers), "m3", "Between {}m and {}m".format(beam_element.n1.x, beam_element.n1.x), False]
        cost_table = [*cost_table, row]
    row = ["Concrete", round(total_concrete_cost, 2), round(volume, 2), "m3", "", True]
    cost_table = [*cost_table, row]
    
    # Longitudinal
    total_long_bar_cost = total_length = 0
    for lb in concrete_beam.long_steel_bars:
        total_long_bar_cost += lb.cost
        total_length += lb.length
        row = ["Longitudinal bar",
            round(lb.cost, decimal_numbers),
            round(lb.length, decimal_numbers),
            "m",
            "Diameter {}mm. Between {}m and {}m".format(abs(lb.diameter*10),
                                                        round(lb.long_begin,decimal_numbers),
                                                        round(lb.long_end,decimal_numbers)),
            False]
        cost_table = [*cost_table, row]
        

    row = ["Longitudinal bar",
        round(total_long_bar_cost, decimal_numbers),
        round(total_length, decimal_numbers),
        "m", "", True]

    cost_table = [*cost_table, row]

    # Transversal Bar
    total_transv_bar_cost = total_length = 0
    for lb in concrete_beam.transv_steel_bars:
        total_transv_bar_cost += lb.cost
        total_length += lb.length
        row = ["Transversal bar",
            round(lb.cost, decimal_numbers),
            round(lb.length, decimal_numbers),
            "m",
            "{}cm x {}cm. Diameter {}mm. Placed in {}m ".format(round(lb.width,decimal_numbers),
                                                                round(lb.height,decimal_numbers),
                                                                abs(lb.diameter*10),
                                                                round(lb.x,decimal_numbers)), False]
        cost_table = [*cost_table, row]
        
    row = ["Transversal bar",
        round(total_transv_bar_cost, decimal_numbers),
        round(total_length, decimal_numbers),
        "m", "", True]

    cost_table = np.array([*cost_table, row])

    is_subtotal = cost_table[:, 5]
    is_subtotal = is_subtotal != "False"
    subtotal_table = cost_table[is_subtotal, :]

    return total_concrete_cost + total_transv_bar_cost + total_long_bar_cost, cost_table, subtotal_table


import numpy as np
import matplotlib.pyplot as plt
import copy
from fconcrete.helpers import getAxis

class TransvSteelBar():
    def __init__(self, x, height, width, diameter, space_after, area, as_per_cm, anchor, length, cost):
        self.x = x
        self.height = height
        self.width = width
        self.diameter = diameter
        self.space_after = space_after
        self.area = area
        self.as_per_cm = as_per_cm
        self.anchor = anchor
        self.length = length
        self.cost =  cost
        
    def __repr__(self):
        return str(self.__dict__)+'\n'
    
    def plot(self, c=2, ax=None, fig=None, color_plot="blue"):
        """
            Plot the transversal vision of the transversal bar.
        """
        height, width, diameter = self.height, self.width, self.diameter
        x0, y0 = -width/2, c
        
        fig, ax = getAxis((x0, diameter), (x0+diameter+width-diameter, c+height)) if ax == None else (fig, ax)
        
        left_bar = plt.Rectangle((x0,y0), diameter, height, hatch="/", fill=False)
        right_bar = plt.Rectangle((width/2-diameter,y0), diameter, height, hatch="/", fill=False)
        top_bar = plt.Rectangle((x0+diameter,y0+height-diameter), width-2*diameter, diameter, hatch="/", fill=False)
        bottom_bar = plt.Rectangle((x0+diameter,y0), width-2*diameter, diameter, hatch="/", fill=False)
        anchor = plt.Rectangle((width/2+diameter/2, y0+height-diameter/2), diameter, self.anchor, hatch="|", fill=False, angle=135) #anchor

        ax.add_artist(left_bar)
        ax.add_artist(right_bar)
        ax.add_artist(top_bar)
        ax.add_artist(bottom_bar)
        ax.add_artist(anchor)

        return fig, ax # if return_ax else None
        

class TransvSteelBars():
    """
        Class that defines a the TransvSteelBar list with easy to work properties and methods.
    """
    def __init__(self, steel_bars=[]):
        self.steel_bars = np.array(steel_bars)
        self.x = np.array([ steel_bar.x for steel_bar in self.steel_bars ])
        self.heights = np.array([ steel_bar.height for steel_bar in self.steel_bars ])
        self.widths = np.array([ steel_bar.width for steel_bar in self.steel_bars ])
        self.diameters = np.array([ steel_bar.diameter for steel_bar in self.steel_bars ])
        self.space_afters = np.array([ steel_bar.space_after for steel_bar in self.steel_bars ])
        self.areas = np.array([ steel_bar.area for steel_bar in self.steel_bars ])
        self.as_per_cms = np.array([ steel_bar.as_per_cm for steel_bar in self.steel_bars ])
        self.lengths = np.array([ steel_bar.length for steel_bar in self.steel_bars ])
        self.costs = np.array([ steel_bar.cost for steel_bar in self.steel_bars ])
        self.length = sum(self.lengths)
        self.cost = sum(self.costs)

    def add(self, new_steel_bars):
        """
            Add a array of Load in the Loads instance.
        """
        previous_steel_bars = self.steel_bars
        if str(type(new_steel_bars)) == "<class 'fconcrete.StructuralConcrete.TransvSteelBar.TransvSteelBar.TransvSteelBars'>":
            concatenation = list(np.concatenate((previous_steel_bars,new_steel_bars.steel_bars)))
            concatenation.sort(key=lambda x: x.long_begin, reverse=False)
            new_steel_bars = np.array(concatenation)
            
        elif str(type(new_steel_bars)) == "<class 'fconcrete.StructuralConcrete.TransvSteelBar.TransvSteelBar.TransvSteelBar'>":
            new_steel_bars = np.append(previous_steel_bars,new_steel_bars)
        self.__init__(new_steel_bars)
    
    def getTransversalBarAfterX(self, x):
        """
            Get the next transversal bar in x or after.
        """
        return self.steel_bars[self.x >= x][0]
    
    def changeProperty(self, prop, function, conditional=lambda x:True):
        """
            Change all properties of the TransvSteelBar in a single function.
        """
        steel_bars = copy.deepcopy(self)
        for previous_steel_bar in steel_bars:
            if conditional(previous_steel_bar):
                current_attribute_value = getattr(previous_steel_bar, prop)
                setattr(previous_steel_bar, prop, function(current_attribute_value)) 
        return steel_bars

    def plotLong(self):
        """
            Plot the longitudinal vision of the transversal bar.
        """
        for x, height in zip(self.x, self.heights):
            plt.plot([x,x], [0,height])
            
    def __getitem__(self, key):
        return self.steel_bars[key]
    
    def __repr__(self):
        return str(self.steel_bars)